/********************************************************************\

  Name:         frontend.c
  Created by:   Stefan Ritt

  Contents:     Frontend code for PCOS+ADC+TDC readout


  Revision history
  ------------------------------------------------------------------
  date         by    modification
  ---------    ---   ------------------------------------------------
  25-JAN-95    SR    created


\********************************************************************/

/* standard include files */
#include <stdio.h>
#include <dos.h>
#include <string.h>
#include <conio.h>

/* midas include file */
#include "midas.h"
#define BYTE_DEFINED

/* include files for hardware access */
#include "hardware.h"
#include "mhyt1331.h"

/* slot number etc */
#include "frontend.h"

/* event structures */
#include "..\event.h"

/*-- System Globals ------------------------------------------------*/

/* The frontend name (client name) as seen by other MIDAS clients   */
char *frontend_name = "Trigger Frontend";

/* frontend_loop is called periodically if this variable is TRUE    */
BOOL frontend_call_loop = FALSE;

/* a frontend status page is displayed with this frequency in ms */
INT display_period = 3000;

/* display function in mfe.c */
extern void display(BOOL bInit);

/*-- Experiment specific Globals -----------------------------------*/

TRIGGER_SETTINGS trigger_settings;
SCALER_SETTINGS  scaler_settings;
EXP_PARAM        exp_param;

/*-- Function declarations -----------------------------------------*/

BOOL poll_trigger_event(INT count, BOOL test);
INT  read_trigger_event(char *pevent);
INT  read_scaler_event(char *pevent);

/*-- Equipment list ------------------------------------------------*/

EQUIPMENT equipment[] = {

  { "Trigger",            /* equipment name */
    1, 0,                 /* event ID, trigger mask */
    "SYSTEM",             /* event buffer */
    EQ_POLLED,            /* equipment type */
    "FIXED",              /* format */
    TRUE,                 /* enabled */
    RO_ODB |              /* write event to odb */
    RO_RUNNING,           /* read only when running */
    100,                  /* poll for 100ms */
    0,                    /* stop run after this event limit */
    FALSE,                /* log history */
    "", "",
    read_trigger_event,   /* readout routine */
    poll_trigger_event    /* polling routine */
  },

  { "Scaler",             /* equipment name */
    2, 0,                 /* event ID, trigger mask */
    "SYSTEM",             /* event buffer */
    EQ_PERIODIC,          /* equipment type */
    "FIXED",              /* format */
    TRUE,                 /* enabled */
    RO_ODB |              /* write event to odb */
    RO_RUNNING |
    RO_TRANSITIONS,       /* read when running and on transitions */
    10000,                /* read every 10 sec */
    0,                    /* stop run after this event limit */
    FALSE,                /* log history */
    "", "",
    read_scaler_event     /* readout routine */
  },

  { "" }
};


/*------------------------------------------------------------------*/

BOOL pcos_init()
{
int   i, n, last_plane, wire_order;
WORD  q, x;
WORD  log_addr, contr_addr;
BYTE  idel, ithres;

  /**** init DBI ****/
  CAMOQ(CS_DBI, 0, 9, 0, &q, &x);
  if (!q || !x)
    {
    printf("Error accessing DBI: Q = %d, X = %d\n", q, x);
    return FALSE;
    }

  n = sizeof(trigger_settings.pcos.slot) / sizeof(WORD);

  last_plane = -1;

  for(i = 0; i < n; i++)
    {
    idel   = (BYTE) min(255, (trigger_settings.pcos.delay[i] - 300) / 1.5);
    ithres = (BYTE) min(255, trigger_settings.pcos.threshold[i]/0.06);

    contr_addr = (1 << 10);

    /**** set logical address as combination of plane and wire order ****/
    log_addr = trigger_settings.pcos.plane[i] << 4;
    if (trigger_settings.pcos.plane[i] != last_plane)
      {
      last_plane = trigger_settings.pcos.plane[i];
      wire_order = 0;
      }
    else
      wire_order++;
    log_addr |= wire_order;
    log_addr &= 0x1FF;

    CAMO(CS_DBI, 0, 16, contr_addr | (1<<3) | (1<<1));
    CAMO(CS_DBI, 0, 16, (1<<15) | log_addr);
    CAMO(CS_DBI, 0, 16, (1<<15) | (trigger_settings.pcos.slot[i]<<3) |
                                    (log_addr>>8));

    /**** set delay ****/
    CAMO(CS_DBI, 0, 16, contr_addr | (1<<3) | (1<<2));
    CAMO(CS_DBI, 0, 16, (1<<15) | idel);
    CAMO(CS_DBI, 0, 16, (1<<15) | (trigger_settings.pcos.slot[i]<<3));

    /**** set threshold ****/
    CAMO(CS_DBI, 0, 16, contr_addr | (1<<3) | (1<<2));
    CAMO(CS_DBI, 0, 16, (1<<15) | ithres);
    CAMO(CS_DBI, 0, 16, (1<<15) | (trigger_settings.pcos.slot[i]<<3) | 1);
    }

  /**** no inhibit any more ****/
  CAMO(CS_DBI, 0, 16, contr_addr);

  return TRUE;
}

/*------------------------------------------------------------------*/

BOOL pcos_read(TRIGGER_EVENT *pevent)
{
WORD  d, n, plane, n_planes, n_max_hit, i, j, cluster_width, wire_pos;
WORD  multiplicity[10];
INT   *pdata[10], pos;

  n = sizeof(trigger_settings.pcos.slot) / sizeof(WORD);
  n_planes = trigger_settings.pcos.plane[n-1] + 1;
  n_max_hit = sizeof(pevent->ax) / sizeof(int);

  for (i=0 ; i < n_planes ; i++) 
    {
    multiplicity[i] = 0;
    pdata[i] = pevent->ax + i*n_max_hit;
    for (j=0 ; j<n_max_hit ; j++)
      pdata[i][j] = -1000;
    }

  cluster_width = 0;

  /**** init memory counter ****/
  CAMO8(CS_DBI, 1, 11, 0);

  /**** read back data from memory ****/
  CAMI(CS_DBI, 0, 2, &d);
  n = d & 0xFFF;

  /**** read out data words ****/
  for (i=0 ; i < n-1 ; i++)
    {
    CAMI(CS_DBI, 0, 2, &d);

    if ((d & 0xC000) == 0xC000)
      break; //delimiter found

    if (d & (1<<15U))
      {
      /**** clusterized data ****/
      cluster_width = max(cluster_width, (d & 0xF));
      CAMI(CS_DBI, 0, 2, &d);
      if ((d & 0xC000) == 0xC000)
        break; //delimiter found
      plane    = (d & 0x7C00)>>10;
      wire_pos = (d & 0x3ff);
      }
    else
      {
      /**** non-clusterized data ****/
      cluster_width = max(cluster_width, 1);
      plane    = (d & 0x7C00)>>10;
      wire_pos = (d & 0x3ff);
      }

    /**** use only position, no cluster width information ****/

    if (plane >= n_planes)
      {
      CAMO(CS_DBI, 0, 11, 0);
      return FALSE;
      }

    /**** convert to mm, plane center = 0 ****/
    pos = (INT) ((wire_pos / 2.0 - trigger_settings.pcos.n_wires[plane]/2.0) * 
                  trigger_settings.pcos.wire_spacing[plane]);
    if (trigger_settings.pcos.mirror[plane])
      pos = -pos;

    if (multiplicity[plane] < n_max_hit)
      pdata[plane][multiplicity[plane]] = pos;

    multiplicity[plane]++;
    }

  /**** clear LeCroy 4299 ****/
  CAMO(CS_DBI, 0, 11, 0);

  /**** return error if not all planes fired ****/
/*
  for (i=0 ; i<n_planes ; i++)
    if (multiplicity[i] == 0)
      return FALSE;
*/
  
  return TRUE;
}

/*------------------------------------------------------------------*/

/********************************************************************\
              Callback routines for system transitions

  These routines are called whenever a system transition like start/
  stop of a run occurs. The routines are called on the following
  occations:

  frontend_init:  When the frontend program is started. This routine
                  should initialize the hardware.

  frontend_exit:  When the frontend program is shut down. Can be used
                  to releas any locked resources like memory, commu-
                  nications ports etc.

  begin_of_run:   When a new run is started. Clear scalers, open
                  rungates, etc.

  end_of_run:     Called on a request to stop a run. Can send final
                  scalers, end-of-run event, and close run gates.

  pause_run:      When a run is paused. Should disable trigger events.

  resume_run:     When a run is resumed. Should enable trigger events.

\********************************************************************/

/*-- Frontend Init -------------------------------------------------*/

INT frontend_init()
{
HNDLE hDB, hKey;

  /**** Open MIDAS online database ****/
  cm_get_experiment_database(&hDB, &hKey);

  /* get experiment parameters */
  if (db_find_key(hDB, 0, "/Experiment/Run Parameters", &hKey) != DB_SUCCESS)
    cm_msg(MERROR, "frontend_init", "Cannot find \"/Experiment/Run Parameters\" tree in ODB");
  else
    if (db_open_record(hDB, hKey, &exp_param, sizeof(exp_param), MODE_READ, NULL) != DB_SUCCESS)
      return -1;

  /* get trigger settings */
  if (db_find_key(hDB, 0, "/Equipment/Trigger/Settings", &hKey) != DB_SUCCESS)
    cm_msg(MERROR, "frontend_init", "Cannot find /Equipment/Trigger/Settings key in ODB.");
  else
    if (db_open_record(hDB, hKey, &trigger_settings,
        sizeof(trigger_settings), MODE_READ, NULL) != DB_SUCCESS)
      return -1;

  /* get scaler settings */
  if (db_find_key(hDB, 0, "/Equipment/Scaler/Settings", &hKey) != DB_SUCCESS)
    cm_msg(MERROR, "frontend_init", "Cannot find /Equipment/Scaler/Settings key in ODB.");
  else
    if (db_open_record(hDB, hKey, &scaler_settings,
        sizeof(scaler_settings), MODE_READ, NULL) != DB_SUCCESS)
      return -1;

  /*-- CAMAC Initialization ----------------------------------------*/
  
  CAMAC_INIT();
  CAMAC_SET_CRATE(3);
  CAMAC_Z_CYCLE();
  CAMAC_C_CYCLE();

  /**** init IO 506 ****/
  CAMO8(CS_IO, 1, 26, 0); // enable auto mode
  CAMO8(CS_IO, 2, 16, 0); // NIM level, int. clock
  CAMO8(CS_IO, 0, 26, 0); // enable LAM Flip-Flop

  /**** PCOS init ****/
  if (!pcos_init(trigger_settings))
    return -1;

  /*----------------------------------------------------------------*/

  /**** clear LAM Flip-Flop ****/
  CAMO(CS_IO, 0, 10, 0);

  /**** reset LAM logic ****/
  CAMO8(CS_IO, 1, 16, 0xF);
  CAMO8(CS_IO, 1, 16, 0);

  return CM_SUCCESS;
}

/*-- Frontend Exit -------------------------------------------------*/

INT frontend_exit()
{
  return CM_SUCCESS;
}

/*-- Begin of Run --------------------------------------------------*/

INT begin_of_run(INT run_number, char *error)
{
  return CM_SUCCESS;
}

/*-- End of Run ----------------------------------------------------*/

INT end_of_run(INT run_number, char *error)
{
  return CM_SUCCESS;
}

/*-- Pause Run -----------------------------------------------------*/

INT pause_run(INT run_number, char *error)
{
  return CM_SUCCESS;
}

/*-- Resuem Run ----------------------------------------------------*/

INT resume_run(INT run_number, char *error)
{
  return CM_SUCCESS;
}

/*-- Frontend Loop -------------------------------------------------*/

INT frontend_loop()
{
  return CM_SUCCESS;
}

/*------------------------------------------------------------------*/

/********************************************************************\

  Readout routines for different events

\********************************************************************/

/*-- Poll Trigger event --------------------------------------------*/

BOOL poll_trigger_event(INT count, BOOL test)
{
INT  i;
BYTE q;

  CAMIQ_RSTART(CS_IO, 0, 8);
  for (i=0 ; i<count ; i++)
    {
    CAMIQ_REPEAT(&q);
    if (q)
      if (!test)
        return TRUE;
    }

  return FALSE;
}

/*-- Read trigger event --------------------------------------------*/

LRS1877_DATA   *lrs_tdc;
LRS1877_HEADER lrs_tdc_header;

INT read_trigger_event(char *pevent)
{
TRIGGER_EVENT *trigger_event;
INT           size, timeout;

  trigger_event = (TRIGGER_EVENT *) pevent;
  size = sizeof(TRIGGER_EVENT);
  
  // signal start of readout
  CAMO8(CS_IO, 1, 16, 1<<7);

  /* wait for ADC Q response */
  timeout = 100;
  CAMI_WAITQ(CS_ADC0, 0, 8, &timeout);
  if (timeout == 0)
    {
    cm_msg(MINFO, "read_trigger_event", "No ADC gate");
    return 0;
    }

  /* ADC readout */
  CAMI_LOOP(CS_ADC0, 1, 2, 0, trigger_event->adc);  // readout
  CAMO8(CS_ADC0, 0, 9, 0);                          // clear

  /* wait for TDC Q response */
  // CAMI_WAITQ(CS_TDC0, 0, 8);

  /* TDC readout */
  CAMI_LOOP(CS_TDC0, 0, 1, 0, trigger_event->tdc);  // readout
  CAMO8(CS_TDC0, 0, 9, 0);                          // clear

  if (!pcos_read(trigger_event) && !exp_param.pedestal_run)
    size = 0;

  // clear LAM Flip-Flop
  CAMO(CS_IO, 0, 10, 0);

  // reset LAM logic
  CAMO8(CS_IO, 1, 16, 0xF);
  CAMO8(CS_IO, 1, 16, 0);

  /**** return event size ****/
  return size;
}

/*-- Scaler event --------------------------------------------------*/

INT read_scaler_event(char *pevent)
{
SCALER_EVENT *scaler_event;
int          i;

  scaler_event = (SCALER_EVENT *) pevent;

  scaler_event->scaler0[1] = 0;

  for (i=scaler_settings.first_channel ;
       i<=scaler_settings.last_channel ; 
       i++)
    CAMI24(CS_SCALER0, i, 2, 
      &scaler_event->scaler0[i-scaler_settings.first_channel]);

  return sizeof(SCALER_EVENT);
}
