/********************************************************************\

  Name:         analyzer.c
  Created by:   Stefan Ritt

  Contents:     PiBeta Analyzer code


  Revision history
  ------------------------------------------------------------------
  date         by    modification
  ---------    ---   ------------------------------------------------
  13-DEC-96    SR    created


\********************************************************************/
                                                        
/*-- Include files -------------------------------------------------*/

/* standard includes */
#include <stdio.h>

/* midas includes */
#include "midas.h"
#include "hardware.h"
#include "..\event.h"

/* cernlib includes */
#define VISUAL_CPLUSPLUS
#include <cfortran.h>
#include <hbook.h>

/*-- Globals -------------------------------------------------------*/

#define N_ADC 2
#define N_TDC 2

/* paw common memory */
#define PAWC_SIZE 1000000  /* increase if necessary */
int PAWC[PAWC_SIZE];

/* paw histo and n-tuple IDs */
#define HISTOID_RAWADC   100
#define HISTOID_CORADC   200
#define HISTOID_TDC      300
#define HISTOID_CHAMBER  400
#define HISTOID_CHAMBER2 500

#define NTID_CALC        1
#define NTID_HIST        2

void get_hbook_filename(char *path);

/* ODB handle defined in mana.c */
extern HNDLE hDB;

/* ODB structures */
extern RUNINFO   runinfo;
extern ANA_PARAM ana_param;
extern EXP_PARAM exp_param;

/* global flag for pedestal run */
BOOL pedestal_run;

/* tag structure */
typedef struct {
  char name[8];
  int  n;
} TAG;

INT book_nt(INT id, char *nt_name, INT n_tag, TAG *taglist, INT n_buffer);

/*-- from trigger events calculated items --------------------------*/

struct {
  float raw_adc[N_ADC];  
  float adc[N_ADC]; 
  float tdc[N_TDC];
  float ax;
  float ay;
  float bx;
  float by;
  float ax2;
  float ay2;
  float bx2;
  float by2;
  float x0;
  float y0;
} calc;

TAG calc_tags[] = {
  "RADC",    N_ADC,
  "ADC",     N_ADC,
  "TDC",     N_TDC,
  "AX",      1,
  "AY",      1,
  "BX",      1,
  "BY",      1,
  "AX2",     1,
  "AY2",     1,
  "BX2",     1,
  "BY2",     1,
  "X0",      1,
  "Y0",      1,
  ""
};

struct {
  float time;  
  float scaler0_0;  
  float scaler0_1;
} hist;

TAG hist_tags[] = {
  "TIME",    1,
  "SC0_0",   1,
  "SC0_1",   1,
  ""
};

/*-- Analyzer init -------------------------------------------------*/

INT ana_init()
{
char name[NAME_LENGTH];
int  i, size;

  /* fetch global flags from ODB */
  pedestal_run = FALSE;
  size = sizeof(pedestal_run);
  db_get_value(hDB, 0, "/experiment/run parameters/pedestal run", 
               &pedestal_run, &size, TID_BOOL);

  /* create global section */
  printf("\n");
  HLIMAP(PAWC_SIZE, "ONLN");
  printf("\n");

  /* book N-tupels */
  book_nt(NTID_CALC, "CALC", sizeof(calc)/sizeof(float), calc_tags, 10000);
  book_nt(NTID_HIST, "HIST", sizeof(hist)/sizeof(float), hist_tags, 100);

  /* define raw adc histos */
  for (i=0 ; i<N_ADC ; i++)
    {
    sprintf(name, "RADC%d", i);
    HBOOK1(HISTOID_RAWADC+i, name, 512, -0.5f, 2047.5f, 0.f);
    }

  /* define calibrated adc histos */
  for (i=0 ; i<N_ADC ; i++)
    {
    sprintf(name, "ADC%d", i);
    HBOOK1(HISTOID_CORADC+i, name, 512, -100.5f, 1947.5f, 0.f);
    }

  /* define TDC histos */
  for (i=0 ; i<N_TDC ; i++)
    {
    sprintf(name, "TDC%d", i);
    HBOOK1(HISTOID_TDC+i, name, 400, -0.25f, 199.75f, 0.f);
    }

  /* define chamber histos */
  HBOOK1(HISTOID_CHAMBER+0, "AX", 128, -128.f, 128.f, 0.f);
  HBOOK1(HISTOID_CHAMBER+1, "AY", 128, -128.f, 128.f, 0.f);
  HBOOK1(HISTOID_CHAMBER+2, "BX", 64, -64.f, 64.f, 0.f);
  HBOOK1(HISTOID_CHAMBER+3, "BY", 64, -64.f, 64.f, 0.f);
  HBOOK1(HISTOID_CHAMBER+4, "X0", 64, -64.f, 64.f, 0.f);
  HBOOK1(HISTOID_CHAMBER+5, "Y0", 64, -64.f, 64.f, 0.f);

  HBOOK2(HISTOID_CHAMBER2+0, "Y0 vs X0", 64, -64.f, 64.f, 64, -64.f, 64.f, 0.f);

  return CM_SUCCESS;
}
        
/*-- Begin of Run --------------------------------------------------*/

FILE *af;

INT ana_begin_of_run(INT run_number, char *error)
{
char filename[256];

  /* clear histos and n-tuples */
  HRESET(0, " ");
  HRESET(NTID_CALC, " ");

  sprintf(filename, "\\midas\\work\\data\\run%04d.asc", run_number);
  af = fopen(filename, "w");
  fprintf(af, "ADC1 ADC2 TDC1 AX AY BX BY AX2 AY2 BX2 AY2\n");

  return CM_SUCCESS;
}

/*-- End of Run ----------------------------------------------------*/

INT ana_end_of_run(INT run_number, char *error)
{
char path[256];

  /* write histos to file if requested */
  get_hbook_filename(path);
  if (path[0])
    HRPUT(0, path, "NT");

  fclose(af);

  return CM_SUCCESS;
}

/*-- Clear histos --------------------------------------------------*/

INT clear_histos(INT id)
{
  /* clear histos and n-tuples */
  HRESET(id, " ");

  return CM_SUCCESS;
}

/*------------------------------------------------------------------*/

/********************************************************************\
  
  Analyzer routines for different events

\********************************************************************/

/*-- Trigger event -------------------------------------------------*/

INT analyze_trigger_event(EVENT_HEADER *pevent)
{
HNDLE          hKey;
static float   adc_sum[N_ADC];
TRIGGER_EVENT  *trigger_event;
int            i;
float          f;

  /* reset calculated values to zero */
  memset(&calc, 0, sizeof(calc));
  trigger_event = (TRIGGER_EVENT *) (pevent+1);

  /*---- copy data form event into calc structure ------------------*/

  for (i=0 ; i<N_ADC ; i++)
    calc.raw_adc[i] = (float) trigger_event->adc[i];

  for (i=0 ; i<N_TDC ; i++)
    calc.tdc[i] = (float) (trigger_event->tdc[i] * 0.1); /* 100ps/bin */

  calc.ax = (float) trigger_event->ax;
  calc.ay = (float) trigger_event->ay;
  calc.bx = (float) trigger_event->bx;
  calc.by = (float) trigger_event->by;
  calc.ax2 = (float) trigger_event->ax2;
  calc.ay2 = (float) trigger_event->ay2;
  calc.bx2 = (float) trigger_event->bx2;
  calc.by2 = (float) trigger_event->by2;

  /*==== code for pedestal run =====================================*/

  if (exp_param.pedestal_run == 1)
    {
    /* init adc_sum on first event */
    if (pevent->serial_number == 1)
      for (i=0 ; i<N_ADC ; i++)
        adc_sum[i] = 0.f;

    /* sum ADC values */
    for (i=0 ; i<N_ADC ; i++)
      adc_sum[i] += calc.raw_adc[i];

    /* calculate pedestals as average of ADCs */
    if (pevent->serial_number % 100 == 1)
      {
      for (i=0 ; i<N_ADC ; i++)
        ana_param.adc_pedestal[i] = adc_sum[i] / 
          (float)pevent->serial_number + 0.5f;

      db_find_key(hDB, 0, "/Analyzer/Parameters", &hKey);
      db_set_record(hDB, hKey, &ana_param, sizeof(ana_param), 0);
      }

    /* finish pedestal run after 5000 events */
    if (pevent->serial_number == 5000)
      cm_transition(TR_STOP, 0, NULL, 0, ASYNC);

    /* don't go on in pedestal run */
    return CM_SUCCESS;    
    }

  /*==== code for data run =========================================*/

  /* ADC pedestal subtraction */
  for (i=0 ; i<N_ADC ; i++)
    calc.adc[i] = calc.raw_adc[i] - ana_param.adc_pedestal[i]; 

  /* fill raw ADC histos */
  for (i=0 ; i<N_ADC ; i++)
    HF1(HISTOID_RAWADC+i, calc.raw_adc[i], 1.f);

  /* fill corrected ADC histos */
  for (i=0 ; i<N_ADC ; i++)
    HF1(HISTOID_CORADC+i, calc.adc[i], 1.f);

  /* fill TDC histos */
  for ( i=0; i<N_TDC; i++ )
    HF1(HISTOID_TDC+i, calc.tdc[i], 1.f);

  /* calculate x0/y0 */
  f = (ana_param.chamber_dist + ana_param.source_dist) / ana_param.chamber_dist;
  calc.x0 = f*(calc.bx - calc.ax) + calc.ax;       
  calc.y0 = f*(calc.by - calc.ay) + calc.ay;       

  /* fill chamber histos */
  HF1(HISTOID_CHAMBER+0, calc.ax, 1.f);
  HF1(HISTOID_CHAMBER+1, calc.ay, 1.f);
  HF1(HISTOID_CHAMBER+2, calc.bx, 1.f);
  HF1(HISTOID_CHAMBER+3, calc.by, 1.f);
  HF1(HISTOID_CHAMBER+4, calc.x0, 1.f);
  HF1(HISTOID_CHAMBER+5, calc.y0, 1.f);

  HF2(HISTOID_CHAMBER2+0, calc.x0, calc.y0, 1.f);

  /* fill calculated values N-tuple */
  HFNOV(NTID_CALC, (float *) &calc);

  /* write ascii file */
  fprintf(af, "%1.0f %1.0f %1.0f %1.0f %1.0f %1.0f %1.0f %1.0f %1.0f %1.0f %1.0f\n", 
               calc.adc[0], calc.adc[1], calc.tdc[0],
               calc.ax, calc.ay, calc.bx, calc.by,
               calc.ax2, calc.ay2, calc.bx2, calc.by2);


  return CM_SUCCESS;
}

/*-- Scaler event --------------------------------------------------*/

extern double sin(double x);

INT analyze_scaler_event(EVENT_HEADER *pevent)
{
SCALER_EVENT *scaler_event;
static DWORD time_offset=0;

  scaler_event = (SCALER_EVENT *) (pevent+1);

  if (time_offset == 0)
    time_offset = ss_time();

  hist.time = (float) (ss_time()-time_offset)/3600.f;
  hist.scaler0_0 = (float) sin(scaler_event->scaler0[0]/3.f);
  hist.scaler0_1 = (float) scaler_event->scaler0[1];

  /* fill history N-tuple */
  HFNOV(NTID_HIST, (float *) &hist);

  return CM_SUCCESS;
}
