/********************************************************************\

  Name:         adccalib.c
  Created by:   Stefan Ritt

  Contents:     PiBeta Analyzer module for ADC calibration.


  Revision history
  ------------------------------------------------------------------
  date         by    modification
  ---------    ---   ------------------------------------------------
  15-APR-97    SR    created
  15-MAY-97    SR    added crystal #28 in exclude list

\********************************************************************/
                                                        
/*-- Include files -------------------------------------------------*/

/* standard includes */
#include <stdio.h>
#include <time.h>
#include "util.h"

/* midas includes */
#include "midas.h"
#include "hardware.h"
#include "event.h"
#include "analyzer.h"

/* cernlib includes */
#ifdef OS_WINNT
#define VISUAL_CPLUSPLUS
#endif

#ifdef __linux__
#define f2cFortran
#endif

#include <cfortran.h>
#include <hbook.h>

/*-- Parameters ----------------------------------------------------*/

ADC_CALIBRATION_PARAM adccalib_param;
extern GLOBAL_PARAM   global_param;
extern EXP_PARAM      exp_param;
extern RUNINFO        runinfo;
extern char           *analyzer_name;

/*-- Local parameters ----------------------------------------------*/

/* ADC groups for common noise supression */
int adc_group[][27] = {
  {  0,  1,  2,  3                                                 },
  {  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15, 16             },
  { 17, 18, 19, 20, 21                                             },
  { 22, 23, 24, 25                                                 },
  { 26, 27, 28, 29, 30, 31                                         },
  { 32, 33, 34, 35, 36, 37, 38, 39                                 },
  { 40, 41, 42, 43                                                 },
  { 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63 },
  { 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99 },
  {100,101,102,103,104,105,106,107,108,109,110,111,112,113,114,115,116 },
  {117,118,119,120,121,122,123,124,125,126,127,128,129,130,131,132,133,134,135,136,137,138,139,140,141,142,143 },
  { 0 },
};

/* CsI crystals "under" plastic vetos */
int pv_csi[N_PVETO] = { 19, 20, 21, 22, 24, 25, 26, 27, 
                        19, 20, 21, 22, 24, 25, 26, 27 };

/*-- Module declaration --------------------------------------------*/

INT adc_calib(char*);
INT adc_calib_init(void);
INT adc_calib_bor(INT run_number);
INT adc_calib_eor(INT run_number);

ADC_CALIBRATION_PARAM_STR(adc_calibration_param_str);

ANA_MODULE adc_calib_module = {
  "ADC calibration",             /* module name           */  
  "Stefan Ritt",                 /* author                */
  adc_calib,                     /* event routine         */
  adc_calib_bor,                 /* BOR routine           */
  adc_calib_eor,                 /* EOR routine           */
  adc_calib_init,                /* init routine          */
  NULL,                          /* exit routine          */
  &adccalib_param,               /* parameter structure   */
  sizeof(adccalib_param),        /* structure size        */
  adc_calibration_param_str,     /* initial parameters    */
};

/*-- temperature event analyzer ------------------------------------*/

double last_temperature = 20;

INT temperature_analyze(char *pevent)
{
TEMPERATURE_EVENT *temp_event;
INT               i;

  temp_event = (TEMPERATURE_EVENT *)pevent;

  /* average over seven temperatures */
  last_temperature = 0;
  for (i=0 ; i<5 ; i++)
    last_temperature += temp_event->input[i];
  last_temperature /= i;

  return SUCCESS;
}

/*-- init routine --------------------------------------------------*/

INT adc_calib_init(void)
{
  /* book histos */
  adc_calib_bor(0);
  return SUCCESS;
}

/*-- BOR routine ---------------------------------------------------*/

#define PED_N_BINS   500
#define PED_X_LOW      0
#define PED_X_HIGH  1000

#define ADC_N_BINS   500
#define ADC_X_LOW      0
#define ADC_X_HIGH  4000

BYTE old_trigger;

INT adc_calib_bor(INT run_number)
{
int    i, size;
char   str[80];
HNDLE  hDB;
static int last_mode = -1;

  if (exp_param.run_mode == 0)
    {
    /* change trigger to clock */
    cm_get_experiment_database(&hDB, NULL);
    size = sizeof(BYTE);
    db_get_value(hDB, 0, "/Equipment/Trigger/Settings/IO506", &old_trigger, &size, TID_BYTE);
    i = (1<<7);
    db_set_value(hDB, 0, "/Equipment/Trigger/Settings/IO506", &i, size, 1, TID_BYTE);

    /* book pedestal histos */
    if (last_mode != exp_param.run_mode)
      for (i=0; i<N_ADC; i++)
        {
        if (HEXIST(ADCCALIB_ID_BASE+i))
          HDELET(ADCCALIB_ID_BASE+i);
        sprintf(str, "Ped%02d", i);
        HBOOK1(ADCCALIB_ID_BASE+i, str, PED_N_BINS, PED_X_LOW.f, PED_X_HIGH.f, 0.f);
        }
    }
  else
    {
    if (last_mode != exp_param.run_mode)
      {
      /* book CADC histos */
      for (i=0; i<N_ADC; i++)
        {
        if (HEXIST(ADCCALIB_ID_BASE+i))
          HDELET(ADCCALIB_ID_BASE+i);
        sprintf(str, "CADC%02d", i);
        HBOOK1(ADCCALIB_ID_BASE+i, str, ADC_N_BINS, ADC_X_LOW.f, ADC_X_HIGH.f, 0.f); 
        }
      /* book pedestal histos */
      for (i=0; i<N_ADC; i++)
        {
        if (HEXIST(ADCCALIB_ID_BASE+500+i))
          HDELET(ADCCALIB_ID_BASE+500+i);
        sprintf(str, "Ped%02d", i);

        if (i<CHN_NAI)
          HBOOK1(ADCCALIB_ID_BASE+500+i, str, 21, -10.5f, 10.5f, 0.f); 
        else
          HBOOK1(ADCCALIB_ID_BASE+500+i, str, 71, -35.5f, 35.5f, 0.f); 
        }
      }
    }

  last_mode = exp_param.run_mode;

  return SUCCESS;
}

/*-- eor routine ---------------------------------------------------*/

INT adc_calib_eor(INT run_number)
{
int    i, i_channel, n_bins, size;
float  x, y[PED_N_BINS], sy, sxy, x_high, x_low;
int    pedestal[N_ADC];
char   str[256];
HNDLE  hDB;
FILE   *f;
time_t now;

  /* pedestal run code */
  if (exp_param.run_mode == 0)
    {
    /* switch back trigger */
    cm_get_experiment_database(&hDB, NULL);
    db_set_value(hDB, 0, "/Equipment/Trigger/Settings/IO506", &old_trigger, sizeof(BYTE), 1, TID_BYTE);

    /*---- analyze pedestal histo ------------------------------------*/

    /* loop through ADC channels */
    for (i_channel=0; i_channel<N_ADC; i_channel++)
      {
      /* move ADC values into array */
      HUNPAK(ADCCALIB_ID_BASE + i_channel, y[0], "HIST", 1);

      /* calculate center of gravity */
      for (i=0,sy=sxy=0.f ; i<PED_N_BINS ; i++)
        {
        x = (float) i/PED_N_BINS*(PED_X_HIGH-PED_X_LOW)+PED_X_LOW;
        sxy += x*y[i];
        sy += y[i];
        }

      pedestal[i_channel] = (int) (sxy/sy + 0.5);
      }

    /* write new pedestal values to online database. */
    cm_get_experiment_database(&hDB, NULL);
    sprintf(str,"/%s/Parameters/ADC calibration/Pedestal", analyzer_name);
    db_set_value(hDB, 0, str, pedestal, sizeof(int)*N_ADC, N_ADC, TID_INT);

    cm_msg(MTALK, "adc_calib_eor", "New pedestals have been written to ODB");
    }
  else
    {
    /* data run code */

    /*---- analyze pedestal histo ------------------------------------*/

    /* get old pedestals */
    cm_get_experiment_database(&hDB, NULL);
    sprintf(str,"/%s/Parameters/ADC calibration/Pedestal", analyzer_name);
    size = sizeof(int)*N_ADC;
    db_get_value(hDB, 0, str, pedestal, &size, TID_INT);

    /* loop through ADC channels */
    for (i_channel=0; i_channel<N_ADC; i_channel++)
      {
      /* move ADC values into array */
      HUNPAK(ADCCALIB_ID_BASE + 500 + i_channel, y[0], "HIST", 1);

      if (i_channel < CHN_NAI)
        {
        n_bins = 21;
        x_low = -10.5f;
        x_high = 10.5f;
        }
      else
        {
        n_bins = 71;
        x_low = -35.5f;
        x_high = 35.5f;
        }

      /* calculate center of gravity */
      for (i=0,sy=sxy=0.f ; i<n_bins ; i++)
        {
        x = (float) i/n_bins*(x_high-x_low)+x_low;
        sxy += x*y[i];
        sy += y[i];
        }

      if (sy > 0)
        pedestal[i_channel] += (int) (sxy/sy + 0.5);
      }

    /* write new pedestal values to online database. */
    db_set_value(hDB, 0, str, pedestal, sizeof(int)*N_ADC, N_ADC, TID_INT);

    /* write pedestals to file */
    if (runinfo.online_mode == 1)
      {
	    f = fopen("c:\\online\\pedestal.txt", "a");
      time(&now);
      strcpy(str, ctime(&now));
      str[24] = 0;
      fprintf(f, "%s\t", str);

      for (i=0 ; i<N_ADC ; i++)
        fprintf(f, "%d\t", pedestal[i]);

      fprintf(f, "\n");
      fclose(f);
      }
    }

  return SUCCESS;
}

/*-- event routine -------------------------------------------------*/

INT adc_calib(char *pevent)
{
INT          n, n_adc;
INT          channel, index, i;
LRS1882_DATA *lrs1882;      
float        *cadc;
float        global_min, min, average;

  /* look for FADC bank, return if not present */
  n_adc = bk_locate(pevent, "FADC", &lrs1882);
  if (n_adc == 0 || n_adc > N_ADC)
    return 1;

  /* create calibrated ADC bank */
  bk_create(pevent, "CADC", TID_FLOAT, &cadc);

  /*---- pedestal run code -----------------------------------------*/

  if (exp_param.run_mode == 0)
    {
    /* fill histos */
    for (i=0 ; i<n_adc ; i++)
      {
      HF1(ADCCALIB_ID_BASE+i, (float) lrs1882[i].data, 1.f);
      *cadc++ = (float) lrs1882[i].data;
      }

    bk_close(pevent, cadc);

    /* stop run after 5000 events */
/*
    if (SERIAL_NUMBER(pevent) == 5000 && runinfo.online_mode)
      cm_transition(TR_STOP, 0, NULL, 0, ASYNC);
*/

    return 1;
    }
  
  /*---- primary pedestal subtraction ------------------------------*/

  /* zero cadc bank */
  for (i=0 ; i<N_ADC ; i++)
    cadc[i] = 0.f;

  /* subtract pedestal */
  for (i=0 ; i<n_adc ; i++)
    cadc[i] = (float) ((double)lrs1882[i].data - adccalib_param.pedestal[i] + 0.5);

  /*---- secondary pedestal subtraction ----------------------------*/
  
  /* find absolute minimum of all CsI */
  global_min = 10000.f;
  for (i=0 ; i<N_CSI ; i++)
    if (cadc[i] < global_min)
      global_min = cadc[i];

  for (index = 0 ; adc_group[index][1] != 0 ; index++)
    {
    /* find minimum value in group */
    min = 10000.f;
    for (i=0 ; adc_group[index][i] != 0 || (index==0 && i==0) ; i++)
      {
      channel = adc_group[index][i];
      if (cadc[channel] < min)
        min = cadc[channel];
      }

    n = 0;
    average = 0.f;
    
    /* If minimum in group is significantly higher than global minimum,
       the group could be completely illiminated. Then we use the global
       average below global minimum to be subtracted from the channels */
    /*
    if (min > global_min + 20)
      {
      for (i=0 ; i<N_CSI ; i++)
        if (cadc[i] < global_min+10)
          {
          average += cadc[i];
          n++;
          }
      }
    else
    */
      {
      /* average channels which are less than 10 counts above minimum in group */
      for (i=0 ; adc_group[index][i] != 0 || (index==0 && i==0) ; i++)
        {
        channel = adc_group[index][i];

        if ((channel < 80 && cadc[channel] < min + 10) || // 10 for CsI et al
            (channel >= 80 && cadc[channel] < min + 30))  // 30 for NaI
          {
          average += cadc[channel];
          n++;
          }
        }
      }

    if (n>0)
      average /= n;

    for (i=0 ; adc_group[index][i] != 0 || i<1 ; i++)
      {
      channel = adc_group[index][i];
      cadc[channel] -= average;
      }
    }

  /* subtract (empty) T0 from B0,D0 and T0 */
  cadc[44] -= cadc[47];
  cadc[45] -= cadc[47];
  cadc[46] -= cadc[47];

  /* subtract (empty) CV0 from TV */
  cadc[64] -= cadc[65];

  /*---- software gain calibration ---------------------------------*/

  for (i=0 ; i<n_adc ; i++)
    cadc[i] *= adccalib_param.software_gain[i];

  /*---- fill histos -----------------------------------------------*/
  
  /* fill ADC histos if above threshold */
  for (i=0 ; i<n_adc ; i++)
    if (cadc[i] > adccalib_param.histo_threshold)
      HF1(ADCCALIB_ID_BASE+i, cadc[i], 1.f);

  /* fill pedestal histos */
  for (i=0 ; i<n_adc ; i++)
    HF1(ADCCALIB_ID_BASE+500+i, cadc[i], 1.f);

  /*---- close calculated bank -------------------------------------*/
  
  bk_close(pevent, cadc+n_adc);
  
  return SUCCESS;
}

