/********************************************************************\

  Name:         adcsum.c
  Created by:   Stefan Ritt

  Contents:     PiBeta Analyzer module for ADC summing


  Revision history
  ------------------------------------------------------------------
  date         by    modification
  ---------    ---   ------------------------------------------------
  04-JUL-97    SR    created

\********************************************************************/
                                                        
/*-- Include files -------------------------------------------------*/

/* standard includes */
#include <stdio.h>
#include <math.h>
#include "util.h"

/* midas includes */
#include "midas.h"
#include "hardware.h"
#include "event.h"
#include "analyzer.h"

/* cernlib includes */
#ifdef OS_WINNT
#define VISUAL_CPLUSPLUS
#endif

#ifdef __linux__
#define f2cFortran
#endif

#include <cfortran.h>
#include <hbook.h>

#ifndef PI
#define PI 3.14159265359
#endif

/*-- Parameters ----------------------------------------------------*/

ADC_SUMMING_PARAM     adc_summing_param;
extern double csi_theta[240];
extern double csi_phi[240];
extern int csi_index[44];

/*-- Static parameters ---------------------------------------------*/

/* ADC channels to exclude in summing */
int adc_exclude[N_ADC];
int adc_exclude_list[] = { -1 };

/* inner crystals */
int csi_inner[] = {
  11, 12, 13, 14, 15, 16, 21, 22, 
  23, 24, 25, 30, 31, 32, -1 };

int csi_outer[] = {
   0,  1,  2,  3,  4,  5,  6,  7,
   8,  9, 10, 17, 18, 19, 20, 26,
  27, 28, 29, 33, 34, 35, 36, 37,
  38, 39, 40, 41, 42, 43, -1 };

/*-- Module declaration --------------------------------------------*/

INT adc_summing(char*);
INT adc_summing_init(void);
INT adc_summing_bor(INT run_number);

ADC_SUMMING_PARAM_STR(adc_summing_param_str);

ANA_MODULE adc_summing_module = {
  "ADC summing",                 /* module name           */  
  "Stefan Ritt",                 /* author                */
  adc_summing,                   /* event routine         */
  NULL,                          /* BOR routine           */
  NULL,                          /* EOR routine           */
  adc_summing_init,              /* init routine          */
  NULL,                          /* exit routine          */
  &adc_summing_param,            /* parameter structure   */
  sizeof(adc_summing_param),     /* structure size        */
  adc_summing_param_str,         /* initial parameters    */
};

/*-- init routine --------------------------------------------------*/

#define CSI_SUM_N_BINS   500
#define CSI_SUM_X_LOW      0
#define CSI_SUM_X_HIGH  4000

#define NAI_SUM_N_BINS   500
#define NAI_SUM_X_LOW      0
#define NAI_SUM_X_HIGH  4000

INT adc_summing_init(void)
{
int  i;
char str[80];

  /* sum histos CsI */
  for (i=0; i<N_CSI; i++)
    {
    sprintf(str, "CsI Sum%02d", i);
    HBOOK1(ADCSUM_ID_BASE+i, str, CSI_SUM_N_BINS, CSI_SUM_X_LOW.f, CSI_SUM_X_HIGH.f, 0.f); 
    }

  return SUCCESS;
}

/*-- event routine -------------------------------------------------*/

INT adc_summing(char *pevent)
{
INT          n_adc,n;
INT          i;
float        *cadc, max_value;
DWORD        *trig;
ASUM_BANK    *asum;


  /* look for CADC bank, return if not present */
  n_adc = bk_locate(pevent, "CADC", &cadc);
  if (n_adc == 0)
    return 1;

  /* look for TRIG bank, return if not present */
  n = bk_locate(pevent, "TRIG", &trig);
  if (n == 0)
    return 1;

  /* create ADC sum bank */
  bk_create(pevent, "ASUM", TID_STRUCT, &asum);

  /*---- CsI ADC summing -------------------------------------------*/

  /* init exclude list */
  /*if (adc_exclude[adc_exclude_list[0]] == FALSE)
    for (i=0 ; adc_exclude_list[i] != -1 ; i++)
      adc_exclude[adc_exclude_list[i]] = TRUE;
  */
  /* sum all channels above threshold */
  asum->csi_sum = 0.f;
  asum->csi_max = -1;
  asum->csi_hit = 0;
  max_value = 0.f;
  for (i=0 ; i<N_CSI ; i++)
    {
    if (/*!adc_exclude[i] &&*/
        cadc[i] >= adc_summing_param.adc_threshold)
      {
      asum->csi_sum += cadc[i];
      asum->csi_hit++;
      }
    if (cadc[i] > max_value)
      {
      max_value = cadc[i];
      asum->csi_max = i;
      }
    }

  /* sum outer/inner */
  asum->csi_inner_sum = asum->csi_outer_sum = 0.f;
  for (i=0 ; csi_inner[i] != -1 ; i++)
    if (cadc[csi_inner[i]] > adc_summing_param.adc_threshold)
      asum->csi_inner_sum += cadc[csi_inner[i]];
  for (i=0 ; csi_outer[i] != -1 ; i++)
    if (cadc[csi_outer[i]] > adc_summing_param.adc_threshold)
      asum->csi_outer_sum += cadc[csi_outer[i]];

  /* fill sum histo */
  if (asum->csi_max >= 0)
    HF1(ADCSUM_ID_BASE+asum->csi_max, asum->csi_sum, 1.f);

  /*---- close calculated bank -------------------------------------*/
  
  bk_close(pevent, asum+1);
  
  return SUCCESS;
}

