/********************************************************************\

  Name:         angle.c
  Created by:   Penny Slocum

  Contents:     PiBeta Analyzer module for calculation of angles of
                incident on CsI swing from wire chambers.


  Revision history
  ------------------------------------------------------------------
  date         by    modification
  ---------    ---   ------------------------------------------------
  29-APR-97    SR    created
  12-MAY-97    PS    added BOR swing rotation and adjusted wc offsets.
  12-JUN-97    PS    added several new variables to ANGL bank, and
                     corrected small bug for cases where adc_sum=0.

\********************************************************************/
                                                        
/*-- Include files -------------------------------------------------*/

/* standard includes */
#include <stdio.h>
#include <math.h>

/* midas includes */
#include "midas.h"
#include "hardware.h"
#include "event.h"
#include "analyzer.h"

/*-- Parameters ----------------------------------------------------*/

extern GLOBAL_PARAM global_param;
extern EXP_PARAM exp_param;
ANGLE_PARAM angle_param;

extern double csi_theta[240];
extern double csi_phi[240];

/*-- Static parameters ---------------------------------------------*/

#ifndef sqr
#define sqr(x) ((x)*(x))
#endif
#ifndef PI
#define PI 3.14159265359
#endif

double sphere_radius   = 260;
double sphere_offset_x = 23.34;
double sphere_offset_y = 0;
double sphere_offset_z = -1120.75;
double dist_chambers   = -702.5;
double out_mat[3][3]; /* calculated in BOR routine */

int csi_index[] = {
  115, 164, 191, 124,  74, 120, 168, 195, 129,  15,  /*  0: 9 */
   65, 170,  73,  24,  20,  70, 174,  78,  29,  16,  /* 10:19 */
  116, 123,  23,   2,  21, 121, 128,  28,  66, 165,  /* 20:29 */ 
   72,  22,  71, 169,  77, 203, 204, 122, 205, 206,  /* 30:39 */
  223, 224, 225, 226 };                              /* 40:43 */

/*-- Module declaration --------------------------------------------*/

INT angle(char*);
INT call_rotate(INT run_number);

ANGLE_PARAM_STR(angle_param_str);

ANA_MODULE angle_module = {
  "Angle",                       /* module name           */  
  "Penny Slocum",                /* author                */
  angle,                         /* event routine         */
  call_rotate,                   /* BOR routine           */
  NULL,                          /* EOR routine           */
  NULL,                          /* init routine          */
  NULL,                          /* exit routine          */
  &angle_param,                  /* parameter structure   */
  sizeof(angle_param),           /* structure size        */
  angle_param_str,               /* initial parameters    */
};

/*-- Rotate matrix -------------------------------------------------*/

void rotate(int axis, double a, double out_mat[3][3])
{
int    i, j, k;
double rot_mat[3][3], in_mat[3][3];

  /* Convert to radians */
  a = a/180*PI;

  for (i=0 ; i<3 ; i++)
    for (j=0 ; j<3 ; j++)
      if (i == j)
        rot_mat[i][j] = 1;
      else
        rot_mat[i][j] = 0;

   if (axis == 0)
    {
    rot_mat[1][1] = cos(a);
    rot_mat[1][2] = -sin(a);
    rot_mat[2][1] = sin(a);
    rot_mat[2][2] = cos(a);
    }
  if (axis == 1)
    {
    rot_mat[0][0] = cos(a);
    rot_mat[0][2] = sin(a);
    rot_mat[2][0] = -sin(a);
    rot_mat[2][2] = cos(a);
    }
  if (axis == 2)
    {
    rot_mat[0][0] = cos(a);
    rot_mat[0][1] = -sin(a);
    rot_mat[1][0] = sin(a);
    rot_mat[1][1] = cos(a);
    }

  /* Place OUT_MAT matrix into IN_MAT matrix, and zero OUT_MAT */
  for (i=0 ; i<3 ; i++)
    for (j=0 ; j<3 ; j++)
      {
      in_mat[i][j] = out_mat[i][j];
      out_mat[i][j] = 0;
      }

  /* Do matrix multiplication */
  for (i=0 ; i<3 ; i++)
    for (k=0 ; k<3 ; k++)
      for (j=0 ; j<3 ; j++)
        out_mat[i][k] = out_mat[i][k] + in_mat[i][j]*rot_mat[j][k];
}

/*-- BOR routine ---------------------------------------------------*/

INT call_rotate(INT run_number)
/* Construct ROTATION MATRIX 'out_mat' for transformation to Geant frame. */
{
int i, j;

  /* First, initialize rotation matrix to "1". */
  for (i=0 ; i<3 ; i++)
    for (j=0 ; j<3 ; j++)
      if (i == j)
        out_mat[i][j] = 1;
      else
        out_mat[i][j] = 0;

  /* Warning:  these rotations happen in the reverse order as entered. */

  /* Rotate the beam about the y-axis. ('theta' in swing lingo.) */
  rotate(1, -180-exp_param.theta_swing, out_mat);

  /* Rotate the beam about the x-axis. ('phi' in swing language) */
  rotate(0, -exp_param.phi_swing, out_mat);

  /* Rotate the beam about the Z-axis for (Geant sys --> lab sys.) */
  rotate(2, -90, out_mat);

  return SUCCESS;
}

/*-- event routine -------------------------------------------------*/

INT angle(char *pevent)
{
float        *cadc;
MWPC_BANK    *mwpc;
ANGL_BANK    *angl;
int          i;
double       xa, ya, xb, yb, x0, y0, z0,
             xk0, yk0, zk0, xk, yk, zk,
             vx, vy, vz, vxk, vyk, vzk, d,
             x, y, z,
             x_l, x_r, y_l, y_r, z_l, z_r, a, b, c,
             sum_array, sum_q_a, theta_q, phi_q, sum_theta_q,
             sum_phi_q, adc_theta, adc_phi, cno_max, adc_max,
             theta_res, phi_res, aoi, phi, theta, ang_res;
double       q[N_CSI];

  /* look for MWPC bank, return if not present */
  if (bk_locate(pevent, "MWPC", &mwpc) == 0)
    return SUCCESS;

  /* look for CADC bank, return if not present */
  if (bk_locate(pevent, "CADC", &cadc) == 0)
    return SUCCESS;

  /* create calculated bank ANGL */
  bk_create(pevent, "ANGL", TID_STRUCT, &angl);

  /*---- perform calculations --------------------------------------*/

  /* convert wire number to lab frame [mm] */
  xa = mwpc->xa - sphere_offset_x;
  ya = -mwpc->ya - sphere_offset_y;
  xb = mwpc->xb - sphere_offset_x;
  yb = mwpc->yb - sphere_offset_y;

  /* Particle vertex in lab frame. (Origin is at center of Chamber 'A'.) */
  x0 = xa;
  y0 = ya;
  z0 = 0.;

  /* Particle vertex in Geant frame. (Origin is at center of sphere.) */
  xk0 = 0;
  yk0 = 0;
  zk0 = 0;

  /* Hit point on sphere in Geant frame. */
  xk = 0.;
  yk = 0.;
  zk = 0.;

  /* Velocity unit vector in lab frame. */
  vx = 0;
  vy = 0;
  vz = 0;

  /* Velocity unit vector in Geant frame. */
  vxk = 0;
  vyk = 0;
  vzk = 0;

  /* Hit point on sphere in lab frame. */
  x = 0;
  y = 0;
  z = 0;

  /* Algebraic roots to geometry calculations. */
  x_l = 0;
  x_r = 0;
  y_l = 0;
  y_r = 0;
  z_l = 0;
  z_r = 0;

  /* Quadratic formula components for geometry calculations. */
  a = 0;
  b = 0;
  c = 0;

  /* Angular resolution calculation variables. */
  sum_array   = 0;
  sum_q_a     = 0;
  theta_q     = 0;
  phi_q       = 0;
  sum_theta_q = 0;
  sum_phi_q   = 0;
  adc_theta   = 0;
  adc_phi     = 0;
  cno_max     = 0;
  adc_max     = 0;
  theta_res   = 0;
  phi_res     = 0;

  for (i=0 ; i<44; i++)
	  q[i] = 0;

  /* wire chamber extrapolation onto sphere */
  if (xa != xb)
    {
    a = sqr(dist_chambers)/sqr(xb-xa) + sqr(yb-ya)/sqr(xb-xa) + 1;
    b = (-2*dist_chambers*sphere_offset_z + 2*ya*(yb-ya))/(xb-xa) + 2*xa;
    c = sqr(sphere_offset_z) - sqr(sphere_radius) + sqr(xa) + sqr(ya);
      
    x_l = (-b + sqrt(sqr(b) - 4*a*c))/(2*a) + xa;
    x_r = (-b - sqrt(sqr(b) - 4*a*c))/(2*a) + xa;
    y_l = (x_l-xa)*(yb-ya)/(xb-xa) + ya;
    y_r = (x_r-xa)*(yb-ya)/(xb-xa) + ya;
    z_l = sqrt(sqr(sphere_radius) - sqr(x_l) - sqr(y_l)) + sphere_offset_z;
    z_r =-sqrt(sqr(sphere_radius) - sqr(x_l) - sqr(y_l)) + sphere_offset_z;

    /* Now choose correct set of algebraic roots. */
	  if (z_l > z_r)
	    z = z_r;
	  else
	    z = z_l;

    if (xb > xa) 
      {
	    if (x_l > x_r) 
        {
		    x = x_l;
		    y = y_l;
	      } 
      else 
        {
		    x = x_r;
		    y = y_r;
	      }
	    } 
    else 
      {
	    if (x_l < x_r) 
        {
		    x = x_l;
		    y = y_l;
	      } 
      else 
        {
		    x = x_r;
		    y = y_r;
	      }
	    }
    }
  else /* xa != xb */
    {
    x = xa;
    if (ya != yb)
      {
      a = 1 + sqr(dist_chambers/(yb-ya));
      b = -2*sphere_offset_z*dist_chambers/(yb-ya) + 2*ya;
      c = sqr(sphere_offset_z) - sqr(sphere_radius) + sqr(x) + sqr(ya);
      y_r = (-b + sqrt(sqr(b) - 4*a*c))/(2*a) + ya;
      y_l = (-b - sqrt(sqr(b) - 4*a*c))/(2*a) + ya;
      z_r = sqrt(sqr(sphere_radius) - sqr(y_r) - sqr(x)) + sphere_offset_z;
      z_l =-sqrt(sqr(sphere_radius) - sqr(y_r) - sqr(x)) + sphere_offset_z;

      /* Again, choose correct points from algebraic roots. */
      if (z_l > z_r)
        z = z_r;
      else
        z = z_l;

      if (xb > xa) 
        {
        if (x_l >  x_r)
          {
          x = x_l;
          y = y_l;
          }
        else
          {
          x = x_r;
          y = y_r;
          }
        }
      else
        {
        if (x_l < x_r) 
          {
          x = x_l;
          y = y_l;
          }
        else
          {
          x = x_r;
          y = y_r;
          }
        }
      }
    else  /* (ya != yb) */
      {
      y = ya;
      z_r =  sqrt(sqr(sphere_radius) - sqr(y) - sqr(x)) + sphere_offset_z;
      z_l = -sqrt(sqr(sphere_radius) - sqr(y) - sqr(x)) + sphere_offset_z;
      if (z_l > z_r)
        z = z_r;
      else
        z = z_l;
      } /* (ya != yb) */
    } /* (xa != xb) */

  /* From above, we now know x,y,z hit points on sphere in lab frame. */

  /* Calculate angle of incidence with detector. */
  aoi = acos( (x*(x-xa) + y*(y-ya) + z*(z-sphere_offset_z)) /
              (sqrt(sqr(x) + sqr(y) + sqr(z-sphere_offset_z)) *
               sqrt( sqr(x-xa) + sqr(y-ya) + sqr(z) )));

  aoi = aoi*180/PI; /* convert aoi to degrees */

  /* Calculate unit vector particle momentum from wc coordinates. */
  vx = (xb-xa);
  vy = (yb-ya);
  vz = (dist_chambers-0);
  d = sqrt(sqr(vx) + sqr(vy) + sqr(vz));
  vx = vx/d;
  vy = vy/d;
  vz = vz/d;

  /* Do TRANSLATION into Ketevi's (geant) coordinate system. */
  z = z - sphere_offset_z;
  z0 = z0 - sphere_offset_z;

  /* Now do the ROTATION, using previously contructed matrix. 
     impact points on sphere in Geant system. */

  xk = out_mat[0][0]*x + out_mat[0][1]*y + out_mat[0][2]*z;
  yk = out_mat[1][0]*x + out_mat[1][1]*y + out_mat[1][2]*z;
  zk = out_mat[2][0]*x + out_mat[2][1]*y + out_mat[2][2]*z;

  /* particle vertex points in Geant system. (at wire chamber A) */
  xk0 = out_mat[0][0]*x0 + out_mat[0][1]*y0 + out_mat[0][2]*z0;
  yk0 = out_mat[1][0]*x0 + out_mat[1][1]*y0 + out_mat[1][2]*z0;
  zk0 = out_mat[2][0]*x0 + out_mat[2][1]*y0 + out_mat[2][2]*z0;

  /* Momentum unit vector in Geant system */
  vxk = out_mat[0][0]*vx + out_mat[0][1]*vy + out_mat[0][2]*vz;
  vyk = out_mat[1][0]*vx + out_mat[1][1]*vy + out_mat[1][2]*vz;
  vzk = out_mat[2][0]*vx + out_mat[2][1]*vy + out_mat[2][2]*vz;
        
  z = z + sphere_offset_z;
  z0 = z0 + sphere_offset_z;

  /* Calculate theta and phi of particle on sphere in Ketevi's frame. */

  /* phi */
  if ((xk != 0) && (yk != 0))
    {
    phi = atan(yk/xk);
    if ((xk < 0) && (yk > 0))
      phi = phi + PI;
    if ((xk < 0) && (yk < 0)) 
      phi = phi + PI;
    if ((xk > 0) && (yk < 0))
      phi = phi + 2*PI;
    }
  else
    {
    if (yk == 0)
      {
      if (xk > 0)
        phi = 0;
      else
        phi = PI;
      }
    else
      {
      if (yk > 0)
        phi = PI/2;
      else
        phi = 3*PI/2;
      }
    }

  /* theta */
  if ((zk != 0) && (sqrt(sqr(xk) + sqr(yk)) != 0))
    {
    theta = atan(sqrt(sqr(xk) + sqr(yk))/zk);
    if (zk < 0)
      theta = theta = PI;
    }
  else
    {
    if (zk == 0)
      {
      if (sqrt(sqr(xk) + sqr(yk)) != 0)
        theta = PI/2;
      else 
        theta = 0;
      }
    else
      {
      if (zk > 0)
        theta = 0;
      else
        theta = PI;
      }
    }

  /* Find coordinates of hit using ADC values. */
  for (i=0 ; i<N_CSI ; i++)
    if (cadc[i] > global_param.adc_threshold)
      {         
      sum_array += cadc[i];
      if (cadc[i] > adc_max)
        {
        cno_max = i;
        adc_max = cadc[i];
        }
      }

  if (adc_max > 0)
    {
    for (i=0 ; i<N_CSI ; i++)
      {
      if (cadc[i] > global_param.adc_threshold)
        {
        q[i] = cadc[i]/sum_array;

        sum_q_a = sum_q_a + pow(q[i], angle_param.alpha);

        theta_q = csi_theta[csi_index[i]];
        phi_q   = csi_phi[csi_index[i]];
      
        sum_theta_q += theta_q*pow(q[i], angle_param.alpha);
        sum_phi_q   += phi_q*pow(q[i], angle_param.alpha);
        }
      }

    adc_theta = sum_theta_q / sum_q_a;
    adc_phi   = sum_phi_q / sum_q_a;

    /* Convert adc angles to radians */
    adc_theta = PI*adc_theta/180;
    adc_phi   = PI*adc_phi/180;

    /* Check argument of acos */
    ang_res = sin(theta)*cos(phi)*sin(adc_theta)*cos(adc_phi) +
              sin(theta)*sin(phi)*sin(adc_theta)*sin(adc_phi) +
              cos(theta)*cos(adc_theta);

    /* Calculate angular resolution */
    if ((ang_res < 1)&&(ang_res>-1.))
      ang_res = acos(ang_res);
    else
      if (ang_res >= 1.)
        ang_res = 0;
      else
        ang_res = PI;
 
    /* convert adc phi/theta to x/y/z */
    angl->x_adc = sphere_radius*sin(adc_theta)*cos(adc_phi);
    angl->y_adc = sphere_radius*sin(adc_theta)*sin(adc_phi);
    angl->z_adc = sphere_radius*cos(adc_theta);
    angl->cos_adc_theta = cos(adc_theta);
    angl->adc_theta_radians = adc_theta;
    angl->adc_phi_radians = adc_phi;
    angl->theta_resolution_cm = sphere_radius/10.*(adc_theta-theta);
    angl->phi_resolution_cm = sphere_radius/10.*(adc_phi-phi);

    /* Convert everything back to degrees */
    ang_res   = ang_res*180/PI;
    theta     = theta*180/PI;
    phi       = phi*180/PI;
    adc_theta = adc_theta*180/PI;
    adc_phi   = adc_phi*180/PI;
    theta_res = fabs(theta-adc_theta);
    phi_res   = fabs(phi-adc_phi);

    angl->angular_resolution = ang_res;
    angl->adc_theta_degrees = adc_theta;
    angl->adc_phi_degrees = adc_phi;
    angl->x_deviations = (xk - angl->x_adc)/10.; 
    angl->y_deviations = (yk - angl->y_adc)/10.; 
    angl->z_deviations = (zk - angl->z_adc)/10.; 

    } /* adc_max > 0 */

  else /* adc_max = 0 */
    {
    angl->x_adc = 1000.;
    angl->y_adc = 1000.;
    angl->z_adc = 1000.;
    angl->adc_theta_degrees=1000.;
    angl->adc_theta_radians=1000.;
    angl->adc_phi_degrees=1000.;
    angl->adc_phi_radians=1000.; 
    angl->angular_resolution=1000.; 
    angl->cos_adc_theta=100.;
    angl->theta_resolution_cm = 1000.;
    angl->phi_resolution_cm = 1000.;
    angl->x_deviations = 1000.;
    angl->y_deviations = 1000.;
    angl->z_deviations = 1000.;
    }
    
    /* Set result values. */
    angl->x = xk;
    angl->y = yk;
    angl->z = zk;
    angl->angle_of_incidence = aoi;
    angl->theta_degrees = theta;
    angl->phi_degrees = phi;


  /* End of Angular Resolution calculation. */

  /*---- close calculated bank -------------------------------------*/
  bk_close(pevent, angl+1);
  
  return SUCCESS;
}
