/********************************************************************\

  Name:         mwpc.c
  Created by:   Stefan Ritt

  Contents:     PiBeta Analyzer module for MWPC calculations


  Revision history
  ------------------------------------------------------------------
  date         by    modification
  ---------    ---   ------------------------------------------------
  29-APR-97    SR    created


\********************************************************************/
                                                        
/*-- Include files -------------------------------------------------*/

/* standard includes */
#include <stdio.h>

/* midas includes */
#include "midas.h"
#include "hardware.h"
#include "event.h"
#include "analyzer.h"

/* cernlib includes */
#ifdef OS_WINNT
#define VISUAL_CPLUSPLUS
#endif

#ifdef __linux__
#define f2cFortran
#endif

#include <cfortran.h>
#include <hbook.h>

/*-- Parameters ----------------------------------------------------*/

MWPC_CALIBRATION_PARAM mwpc_calib_param;
extern TRIGGER_SETTINGS trigger_settings;

/*-- Static parameters ---------------------------------------------*/


/*-- Module declaration --------------------------------------------*/

INT mwpc_calib(char*);
INT mwpc_init(void);
INT mwpc_bor(INT run_number);

MWPC_CALIBRATION_PARAM_STR(mwpc_calibration_param_str);

ANA_MODULE mwpc_calib_module = {
  "MWPC Calibration",            /* module name           */  
  "Stefan Ritt",                 /* author                */
  mwpc_calib,                    /* event routine         */
  NULL,                          /* BOR routine           */
  NULL,                          /* EOR routine           */
  mwpc_init,                     /* init routine          */
  NULL,                          /* exit routine          */
  &mwpc_calib_param,             /* parameter structure   */
  sizeof(mwpc_calib_param),      /* structure size        */
  mwpc_calibration_param_str,    /* initial parameters    */
};

/*-- init routine --------------------------------------------------*/

INT mwpc_init(void)
{
  /* 2D histos */
  HBOOK2(MWPC_ID_BASE+0, "Chamber A", 64, -128.f, 128.f, 64, -128.f, 128.f, 0.f); 
  HBOOK2(MWPC_ID_BASE+1, "Chamber B", 64, -64.f, 64.f, 64, -64.f, 64.f, 0.f); 
  HBOOK2(MWPC_ID_BASE+2, "Surface projection", 64, -64.f, 64.f, 64, -64.f, 64.f, 0.f); 
  HBOOK2(MWPC_ID_BASE+7, "Bx-Ax vs Bx", 64, -128.f, 128.f, 64, -64.f, 64.f, 0.f); 
  HBOOK2(MWPC_ID_BASE+8, "By-Ay vs By", 64, -128.f, 128.f, 64, -64.f, 64.f, 0.f); 
  /* 1D histo */
  HBOOK1(MWPC_ID_BASE+3, "AX", 64, -128.f, 128.f, 0.f); 
  HBOOK1(MWPC_ID_BASE+4, "AY", 64, -128.f, 128.f, 0.f); 
  HBOOK1(MWPC_ID_BASE+5, "BX", 64, -64.f, 64.f, 0.f); 
  HBOOK1(MWPC_ID_BASE+6, "BY", 64, -64.f, 64.f, 0.f); 

  return SUCCESS;
}

/*-- event routine -------------------------------------------------*/

INT mwpc_calib(char *pevent)
{
INT          i, n, n_pcos, n_planes;
WORD         d, plane, cluster_width, wire_pos;
WORD         multiplicity[10];
float        *pdata, pos, f;
MWPC_BANK    *mwpc;
WORD         *pcos;

  /* look for PCOS bank, return if not present */
  n_pcos = bk_locate(pevent, "PCOS", &pcos);
  if (n_pcos == 0)
    return 1;

  /* create calibrated MWPC bank */
  bk_create(pevent, "MWPC", TID_STRUCT, &mwpc);

  n = sizeof(trigger_settings.pcos.slot) / sizeof(WORD);
  n_planes = trigger_settings.pcos.plane[n-1] + 1;
  pdata = &mwpc->xa;

  for (i=0 ; i < n_planes ; i++) 
    {
    multiplicity[i] = 0;
    pdata[i] = -1000.f;
    }

  /*---- loop through PCOS data ------------------------------------*/

  /* got some strange PCOS banks in the past */
  if (n_pcos > 100)
    return 1;

  for (i=0 ; i < n_pcos ; i++)
    {
    d = pcos[i];
    if (d & (1<<15))
      {
      /* clusterized data */
      cluster_width = (d & 0xF);
      d = pcos[++i];
      plane    = (d & 0x7C00)>>10;
      wire_pos = (d & 0x3ff);
      }
    else
      {
      /* non-clusterized data */
      cluster_width = 1;
      plane    = (d & 0x7C00)>>10;
      wire_pos = (d & 0x3ff);
      }

    if (plane < n_planes)
      {
      /* convert to mm, plane center = 0 */
      pos = (float) ((wire_pos / 2.0 - trigger_settings.pcos.n_wires[plane]/2.0) * 
                    trigger_settings.pcos.wire_spacing[plane]);
      if (trigger_settings.pcos.mirror[plane])
        pos = -pos;

      pdata[plane] = pos;
      multiplicity[plane]++;
      }
    }

  /*  shift spectra towards center */
  mwpc->xa+=(float)6.6;
  mwpc->ya-=(float)17.8;
  mwpc->xb+=(float)11.6;
  mwpc->yb-=(float)23.6;

  /* calculate surface projection */
  f = (mwpc_calib_param.distance_a_b+mwpc_calib_param.distance_b_target) /
       mwpc_calib_param.distance_a_b;
  mwpc->xtgt = f * (mwpc->xb - mwpc->xa) + mwpc->xa;
  mwpc->ytgt = f * (mwpc->yb - mwpc->ya) + mwpc->ya;

  bk_close(pevent, mwpc+1);

  /*---- fill histos -----------------------------------------------*/

  /* x/y plots */
  if (multiplicity[0] == 1 && multiplicity[1] == 1 &&
      multiplicity[2] == 1 && multiplicity[3] == 1)
    {
    HF2(MWPC_ID_BASE+0, mwpc->xa, mwpc->ya, 1.f);
    HF2(MWPC_ID_BASE+1, mwpc->xb, mwpc->yb, 1.f);
    HF2(MWPC_ID_BASE+2, mwpc->xtgt, mwpc->ytgt, 1.f);

    /* phase space plots */
    HF2(MWPC_ID_BASE+7, mwpc->xb, mwpc->xb-mwpc->xa, 1.f);
    HF2(MWPC_ID_BASE+8, mwpc->yb, mwpc->yb-mwpc->ya, 1.f);
    }

  /* 1D histos */
  if (multiplicity[0] == 1)
    HF1(MWPC_ID_BASE+3, mwpc->xa, 1.f);            
  if (multiplicity[1] == 1)
    HF1(MWPC_ID_BASE+4, mwpc->ya, 1.f);
  if (multiplicity[2] == 1)
    HF1(MWPC_ID_BASE+5, mwpc->xb, 1.f);
  if (multiplicity[3] == 1)
    HF1(MWPC_ID_BASE+6, mwpc->yb, 1.f);
  

  return SUCCESS;
}
