/********************************************************************\

  Name:         scaler.c
  Created by:   Stefan Ritt

  Contents:     PiBeta Scaler Analyzer


  Revision history
  ------------------------------------------------------------------
  date         by    modification
  ---------    ---   ------------------------------------------------
  23-APR-97    SR    created


\********************************************************************/
                                                        
/*-- Include files -------------------------------------------------*/

/* standard includes */
#include <stdio.h>
#include <time.h>

/* midas includes */
#include "midas.h"
#include "event.h"

extern RUNINFO          runinfo;
extern TRIGGER_SETTINGS trigger_settings;

/*-- Module declaration --------------------------------------------*/

INT scaler_accum(char*);
INT scaler_clear(INT run_number);
INT scaler_eor(INT run_number);

ANA_MODULE scaler_accum_module = {
  "Scaler accumulation",         /* module name           */  
  "Stefan Ritt",                 /* author                */
  scaler_accum,                  /* event routine         */
  scaler_clear,                  /* BOR routine           */
  scaler_eor,                    /* EOR routine           */
  NULL,                          /* init routine          */
  NULL,                          /* exit routine          */
  NULL,                          /* parameter structure   */
  0,                             /* structure size        */
  NULL,                          /* initial parameters    */
};

/*-- accumulated scalers -------------------------------------------*/

double scaler[400];
double nims[40];
RTES_BANK rates;

/*-- BOR routine ---------------------------------------------------*/

INT scaler_clear(INT run_number)
{
  memset(scaler, 0, sizeof(scaler));
  memset(nims, 0, sizeof(nims));
  return SUCCESS;
}

/*-- EOR routine ---------------------------------------------------*/

INT scaler_eor(INT run_number)
{
FILE   *f;
HNDLE  hDB;
int    size;
double delta;
float  temperature[11];
time_t now;
char   str[80];

  time(&now);
  strcpy(str, ctime(&now));
  str[19] = 0;

  cm_get_experiment_database(&hDB, NULL);
  size = sizeof(temperature);
  db_get_value(hDB, 0, "/Equipment/Temperature/Variables/Input", temperature, &size, TID_FLOAT);
  
  /* write accumulated rates in tabular format */
  f = fopen("e:\\rates_table.txt", "a");
  if (f==NULL)
    cm_msg(MERROR, "scaler_eor", "Cannot open rates.txt");
  else
    {
    /* accumulated seconds */
    delta = nims[11]/100;

    if (delta > 0)
      {
      /* normalize to 1.5 mA beam */
      delta *= (nims[2]/100000/delta) / 1.5;
      if (delta > 0)
        {
        fprintf(f, "RUN%d\t%s\t", runinfo.run_number, str+11);
        fprintf(f, "IO506\t%d\t", trigger_settings.io506);
        fprintf(f, "CsI UL\t%1.2f\t", trigger_settings.da500_hi);
        fprintf(f, "CsI LL\t%1.2f\t", trigger_settings.da500_lo);
        fprintf(f, "Trig raw\t%1.1lf\t", nims[0]/delta);
        fprintf(f, "Trig acc\t%1.1lf\t", nims[1]/delta);
        fprintf(f, "MHC\t%1.3lf\t", nims[2]/100000/nims[11]*100);
        fprintf(f, "B0\t%1.1lf\t", scaler[44]/delta);
        fprintf(f, "B1\t%1.1lf\t", scaler[45]/delta);
        fprintf(f, "B0*B1\t%1.1lf\t", nims[8]/delta);
        fprintf(f, "CsI hi\t%1.1lf\t", nims[9]/delta);
        fprintf(f, "CsI lo\t%1.1lf\t", nims[10]/delta);
        fprintf(f, "Nai inner hi\t%1.1lf\t", nims[6]/delta);
        fprintf(f, "Nai inner lo\t%1.1lf\t", nims[4]/delta);
        fprintf(f, "Nai outer hi\t%1.1lf\t", nims[7]/delta);
        fprintf(f, "Nai outer lo\t%1.1lf\t", nims[5]/delta);
        fprintf(f, "Temp.\t%1.2f\n", temperature[2]);
        }
      }

    fclose(f);
    }

  return SUCCESS;
}

/*-- event routine -------------------------------------------------*/

INT scaler_accum(char *pevent)
{
INT       n, i;
DWORD     *psclr, *psclrn;
double    *pacum, delta;
RTES_BANK *rtes;
FILE      *f;
HNDLE     hDB;
int       size;
float     temperature[11];
time_t    now;
char      str[80];

  /* look for SCLR bank */
  n = bk_locate(pevent, "SCLR", &psclr);
  if (n == 0)
    return 1;

  /* create acummulated scaler bank */
  bk_create(pevent, "ACUM", TID_DOUBLE, &pacum);

  /* accumulate scalers */
  for (i=0 ; i<n ; i++)
    {
    scaler[i] += psclr[i];
    pacum[i] = scaler[i];
    }
  bk_close(pevent, pacum+n);

  /* look for NIMS bank */
  n = bk_locate(pevent, "NIMS", &psclrn);
  if (n == 0)
    return 1;

  /* create acummulated scaler bank */
  bk_create(pevent, "ACUN", TID_DOUBLE, &pacum);

  /* accumulate scalers */
  for (i=0 ; i<n ; i++)
    {
    nims[i] += psclrn[i];
    pacum[i] = nims[i];
    }
  bk_close(pevent, pacum+n);

  // printf("%ld -> %lf\n", psclrn[11], nims[11]);

  /* create rates bank */
  delta = psclrn[11]/100.0;

  /* calculate rates if enough statistics */
  if (delta > 9)
    {
    rates.trigger = psclrn[0] / delta;
    rates.trigger_accepted = psclrn[1] / delta;
    rates.mhc = psclrn[2] / 100000 / delta;
    rates.b0 = psclr[44] / delta;
    rates.b1 = psclr[45] / delta;
    rates.b0_b1 = psclrn[8] / delta;
    rates.csi_hi = psclrn[9] / delta;
    rates.csi_lo = psclrn[10] / delta;
    rates.nai_inner_hi = psclrn[6] / delta;
    rates.nai_inner_lo = psclrn[4] / delta;
    rates.nai_outer_hi = psclrn[7] / delta;
    rates.nai_outer_lo = psclrn[5] / delta;

    /* put rates in bank */
    bk_create(pevent, "RTES", TID_STRUCT, &rtes);
    memcpy(rtes, &rates, sizeof(RTES_BANK));
    bk_close(pevent, rtes+1);

    /* write rates to file */
    time(&now);
    strcpy(str, ctime(&now));
    str[19] = 0;

    cm_get_experiment_database(&hDB, NULL);
    size = sizeof(temperature);
    db_get_value(hDB, 0, "/Equipment/Temperature/Variables/Input", temperature, &size, TID_FLOAT);
  
    f = fopen("e:\\rates.txt", "w");
    if (f==NULL)
      cm_msg(MERROR, "scaler_eor", "Cannot open rates.txt");
    else
      {
      fprintf(f, "RUN%3d         %s\n", runinfo.run_number, str+11);
      fprintf(f, "IO506          %d\n", trigger_settings.io506);
      fprintf(f, "CsI UL         %1.2f V\n", trigger_settings.da500_hi);
      fprintf(f, "CsI LL         %1.2f V\n", trigger_settings.da500_lo);
      fprintf(f, "Trig raw       %1.1lf\n", rates.trigger);
      fprintf(f, "Trig acc       %1.1lf\n", rates.trigger_accepted);
      fprintf(f, "MHC            %1.3lf mA\n", rates.mhc);
      fprintf(f, "B0             %1.1lf\n", rates.b0);
      fprintf(f, "B1             %1.1lf\n", rates.b1);
      fprintf(f, "B0*B1          %1.1lf\n", rates.b0_b1);
      fprintf(f, "CsI hi         %1.1lf\n", rates.csi_hi);
      fprintf(f, "CsI lo         %1.1lf\n", rates.csi_lo);
      fprintf(f, "Nai inner hi   %1.1lf\n", rates.nai_inner_hi);
      fprintf(f, "Nai inner lo   %1.1lf\n", rates.nai_inner_lo);
      fprintf(f, "Nai outer hi   %1.1lf\n", rates.nai_outer_hi);
      fprintf(f, "Nai outer lo   %1.1lf\n", rates.nai_outer_lo);
      fprintf(f, "Temperature    %1.2f\n", temperature[2]);

      fclose(f);
      }
    }

  return SUCCESS;
}
