/********************************************************************\

  Name:         timing.c
  Created by:   Stefan Ritt

  Contents:     PiBeta Analyzer module for trigger timing


  Revision history
  ------------------------------------------------------------------
  date         by    modification
  ---------    ---   ------------------------------------------------
  07-JUL-97    SR    created

\********************************************************************/
                                                        
/*-- Include files -------------------------------------------------*/

/* standard includes */
#include <stdio.h>
#include "util.h"

/* midas includes */
#include "midas.h"
#include "hardware.h"
#include "event.h"
#include "analyzer.h"

/* cernlib includes */
#ifdef OS_WINNT
#define VISUAL_CPLUSPLUS
#endif

#ifdef __linux__
#define f2cFortran
#endif

#include <cfortran.h>
#include <hbook.h>

/*-- Parameters ----------------------------------------------------*/

extern EXP_PARAM      exp_param;

/*-- Static parameters ---------------------------------------------*/

#define TIMING_N_BINS   400
#define TIMING_X_LOW      0
#define TIMING_X_HIGH   200

#define TIME_NAI

#ifdef TIME_NAI
#define N_CHN         N_NAI
#define CHN_OFFSET  CHN_NAI
#else
#define N_CHN     (N_CSI-4)
#define CHN_OFFSET  CHN_CSI
#endif

/* channels to exclude from evaluation */
int tdc_exclude[240];
int tdc_exclude_list[] = { -1 };

/*-- Module declaration --------------------------------------------*/

INT timing(char*);
INT timing_bor(INT run_number);
INT timing_eor(INT run_number);

ANA_MODULE timing_module = {
  "Timing",                      /* module name           */  
  "Stefan Ritt",                 /* author                */
  timing,                        /* event routine         */
  timing_bor,                    /* BOR routine           */
  timing_eor,                    /* EOR routine           */
  NULL,                          /* init routine          */
  NULL,                          /* exit routine          */
  NULL,                          /* parameter structure   */
  0,                             /* structure size        */
  NULL,                          /* initial parameters    */
};

/*-- BOR routine ---------------------------------------------------*/

WORD old_trigger;

INT timing_bor(INT run_number)
{
int   i;
char  str[80];

  if (exp_param.run_mode == 2)
    {
    /* change trigger to prompts */
    /*
    cm_get_experiment_database(&hDB, NULL);
    size = sizeof(BYTE);
    db_get_value(hDB, 0, "/Equipment/Trigger/Settings/IO506", &old_trigger, &size, TID_BYTE);
    i = (1<<4);
    db_set_value(hDB, 0, "/Equipment/Trigger/Settings/IO506", &i, size, 1, TID_BYTE);
    */

    /* book target TDC histos */
    for (i=0; i<N_CHN; i++)
      {
      if (HEXIST(TIMING_ID_BASE+i))
        HDELET(TIMING_ID_BASE+i);
      sprintf(str, "Time%02d", i);
      HBOOK1(TIMING_ID_BASE+i, str, TIMING_N_BINS, TIMING_X_LOW.f, 
             TIMING_X_HIGH.f, 0.f); 
      }
    }

  return SUCCESS;
}

/*-- EOR routine ---------------------------------------------------*/

INT timing_eor(INT run_number)
{
int   i, j, i_channel;
float y[TIMING_N_BINS], y1[TIMING_N_BINS];
float min, max;
float timing[240];
BOOL  valid[240];
FILE  *f;

  if (exp_param.run_mode != 2)
    return SUCCESS;

  /* switch back trigger */
  /*
  cm_get_experiment_database(&hDB, NULL);
  db_set_value(hDB, 0, "/Equipment/Trigger/Settings/IO506", &old_trigger, sizeof(BYTE), 1, TID_BYTE);
  */

  /* init exclude list */
  for (i=0 ; tdc_exclude_list[i] != -1 ; i++)
    tdc_exclude[tdc_exclude_list[i]] = TRUE;

  for (i_channel=0; i_channel<N_CHN; i_channel++)
    {
    if (tdc_exclude[i_channel])
      continue;

    HUNPAK(TIMING_ID_BASE + i_channel, y[0], "HIST", 1);

    /* smooth */
    for (j=0 ; j<4 ; j++)
      {
      y1[0] = y1[TIMING_N_BINS-1] = 0.f;
      for (i=1 ; i<TIMING_N_BINS-1 ; i++)
        y1[i] = (2*y[i] + y[i-1] + y[i+1])/4;
      for (i=0 ; i<TIMING_N_BINS ; i++)
        y[i] = y1[i];
      }

    /* search maximum */
    max = 0.f;
    for (i=0 ; i<TIMING_N_BINS ; i++)
      if (y[i] > max)
        {
        max = y[i];
        timing[i_channel] = (float) i/TIMING_N_BINS*
          (TIMING_X_HIGH-TIMING_X_LOW)+TIMING_X_LOW;
        }
    valid[i_channel] = (max >= 5);
    }

  /* find extrema of all channels */
  min = max = timing[0];
  for (i=0; i<N_CHN; i++)
    {
    if (!valid[i] || tdc_exclude[i])
      continue;
    if (timing[i] < min)
      min = timing[i];
    if (timing[i] > max)
      max = timing[i];
    }

  /* calculate deviations to earliest peak */
  for (i=0; i<N_CHN; i++)
    timing[i] -= min;
  
  /* write file */
  f = fopen("timing.txt", "w");
  for (i=0; i<N_CHN; i++)
    if (tdc_exclude[i])
      fprintf(f, "Chn %3d excluded\n", i);
    else if (valid[i])
      fprintf(f, "Chn %3d needs %3.1f ns of delay\n", i, timing[i]);
    else
      fprintf(f, "Chn %3d does not contain enough hits\n", i);
  fclose(f);

  cm_msg(MTALK, "timing_eor", "Trigger timing written to TIMING.TXT");

  return SUCCESS;
}

/*-- event routine -------------------------------------------------*/

INT timing(char *pevent)
{
float        *cadc, *thit;
ASUM_BANK    *asum;

  /* locate needed banks */
  if (!bk_locate(pevent, "CADC", &cadc) ||
      !bk_locate(pevent, "THIT", &thit) ||
      !bk_locate(pevent, "ASUM", &asum))
    return 1;

  /* only if in TDC calibration mode */
  if (exp_param.run_mode == 2)
    {
    /* if max ADC contains > 70%, fill TDC histo of that channel with B1 TDC */
#ifdef TIME_NAI
    if (cadc[asum->nai_max+CHN_OFFSET] > 0.7*asum->nai_sum && thit[CHN_B1] > 0 
        && cadc[64] < 40.)
      HF1(TIMING_ID_BASE+asum->nai_max, thit[CHN_B1], 1.f);
#else
    if (cadc[asum->csi_max+CHN_OFFSET] > 0.7*asum->csi_sum && thit[CHN_B1] > 0)
      HF1(TIMING_ID_BASE+asum->csi_max, thit[CHN_B1], 1.f);
#endif
    }

  return SUCCESS;
}

