/********************************************************************\

  Name:         util.c
  Created by:   Stefan Ritt

  Contents:     General purpose utility functions for use in the
                PiBeta analyzer


  Revision history
  ------------------------------------------------------------------
  date         by    modification
  ---------    ---   ------------------------------------------------
  03-JUL-97    SR    created


\********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/*----------------------------------------------------------------*/

/*******************************************************************
  
    Marquardt fit. N-dimensional fit to a non-linear function.

*******************************************************************/

/* maximal number of parameters */

#define MAX_PAR 10

/*----------------------------------------------------------------*/

void fderiv(double x, double param[], 
            double (*func)(double x, double param[]), 
            int n_terms, double deriv[])
/******************************************************************\

  calculation of derivations of a function "func" at point x
  in respect to its parameters param[]

\******************************************************************/
{ 
int    i;
double delta, yf;

  yf = func(x, param);

  for (i=0 ; i<n_terms ; i++)
    {
    delta = fabs(param[i])/1000;
    if (delta < 1E-12) 
      delta=1E-12;

	  param[i] += delta;
	  deriv[i] = (func(x, param) - yf) / delta;
	  param[i] -=delta;
    }
}

/*----------------------------------------------------------------------*/

double calc_chisqr(double y[], double yfit[], double weight[], 
                   int n_points, int n_free)
/******************************************************************\
  calculates chisquare for a fit yfit[] in respect to a sample 
  distribution y[]
  
  Input parameters: y[]        sample
                    yfit[]     fit
                    weight[]   statistical weights 
                    n_points   dimension of y,yfit
                    n_free     degree of freedom

  Return value:     chisquare 
\******************************************************************/
{ 
double chisq;
int    i;

  for (i=0,chisq=0 ; i<n_points ; i++)
    chisq += weight[i]*(y[i]-yfit[i])*(y[i]-yfit[i]);

  return(chisq/(double)n_free);
}

/*--------------------------------------------------------------------*/

int matinv(double array[MAX_PAR][MAX_PAR], int n_order)
/******************************************************************\

  inverts matrix array of order n_order
  returns 1 if successful, -1 if matrix is singular

\******************************************************************/
{ 
double amax,save;
int    ik[MAX_PAR],jk[MAX_PAR],k,i,j,l;

  for (k=0;k<n_order;k++)
    {
    amax=0;
label:
    for (i=k;i<n_order;i++) 
      {
      for (j=k;j<n_order;j++)
        {
	      if (fabs(amax) <= fabs(array[i][j]))
          {
	        amax=array[i][j];
	        ik[k]=i;
	        jk[k]=j;
          }
        }
      }

    if (amax==0)
      return -1;

    i=ik[k];
    if ((i-k)<0) 
      goto label;
    else if((i-k)>0)
      {
      for(j=0;j<n_order;j++)
        {
	      save=array[k][j];
	      array[k][j]=array[i][j];
	      array[i][j]=-save;
        }
      }
    j=jk[k];
    if ((j-k)<0) goto label;
    else if ((j-k)>0)
      {
      for (i=0;i<n_order;i++)
        {
	      save=array[i][k];
	      array[i][k]=array[i][j];
	      array[i][j]=-save;
        }
      }

    for (i=0 ; i<n_order ; i++)
      if (i!=k) 
        array[i][k] /= -amax;
    
    for (i=0 ; i<n_order ; i++)
      for (j=0 ; j<n_order ; j++)
	      if (i!=k && j!=k)
	        array[i][j] += array[i][k]*array[k][j];

    for(j=0;j<n_order;j++)
      if(j!=k) array[k][j]/=amax;
    array[k][k]=1/amax;
    }
  
  for (l=0 ; l<n_order ; l++)
    {
    k=n_order-l-1;
    j=ik[k];
    if (j>k)
      for (i=0;i<n_order;i++)
        {
	      save=array[i][k];
	      array[i][k]=-array[i][j];
	      array[i][j]=save;
        }

    i=jk[k];
    if (i>k)
      for (j=0;j<n_order;j++)
        {
	      save=array[k][j];
	      array[k][j]=-array[i][j];
	      array[i][j]=save;
        }
    }

  return(1);
}

/*----------------------------------------------------------------*/

int marquardt_fit(double x[],  double y[], double weight[], 
                  double (*func)(double x, double param[]), 
                  int n_points, int n_terms,
                  double a[],  double sigmaa[],
                  double *flamda,
		              double yfit[], double *chisqr)
/******************************************************************\
  
    calculates a new fit contained in yfit[] to a sample
    distribution y[] by variation of the parameters a[]
    of the fit function; it also calculates the new
    parameters a[] and  the error of the parameters sigmaa[]

    Input  :  x[] array containing x-values of sample
              y[] array containing sample
              weight[] array containing weights
              n_points: number of sample points to fit
              n_terms: number of free parameters
              a[]: array containing parameter values
              flamda: marquart faktor                    

    Output:   a[] new parameters                        
              sigmaa[] errors in parameters             
              yfit[] new fit                            
              chisqr new chisquare

    Returns: 1 on success 
             -1 if variation of parameters doesn't change chi^2

\******************************************************************/
{ 
double alpha[MAX_PAR][MAX_PAR], beta[MAX_PAR], deriv[MAX_PAR];
double array[MAX_PAR][MAX_PAR], b[MAX_PAR], chisq1, buf;
int    i,j,k;

  /* evaluate alpha and beta matrices */
  for (i=0 ; i<n_terms ; i++)
    {
    beta[i]=0;
    for (j=0;j<=i;j++)
      alpha[i][j]=0;
    }

  for (i=0 ; i<n_points ; i++)
    {
    fderiv(x[i], a, func, n_terms, deriv);

    for (j=0 ; j<n_terms; j++)
      {
      beta[j] += weight[i]*(y[i]-yfit[i])*deriv[j];
      for (k=0 ; k<=j ; k++)
	      alpha[j][k] += weight[i]*deriv[j]*deriv[k];
      }
    }

  for (i=0 ; i<n_terms ; i++)
    for (j=0 ; j<i ; j++)
      alpha[j][i]=alpha[i][j];

  chisq1 = *chisqr;

  /* invert modified curvature matrix to find new parameters */
  do
    {
    for (j=0 ; j<n_terms ; j++)
      {
      for (k=0 ; k<n_terms ; k++)
        {
	      buf = sqrt(alpha[j][j] * alpha[k][k]);
	      if (buf == 0)
	        return -1;

	      array[j][k] = alpha[j][k]/buf;
        }
      array[j][j] = 1 + *flamda;
      }

    if (matinv(array, n_terms) == -1)
      return -1;
    
    for (j=0 ; j<n_terms ; j++)
      {
      b[j]=a[j];
      for (k=0;k<n_terms;k++)
	      b[j] += beta[k]*array[j][k] / sqrt(alpha[j][j]*alpha[k][k]);
      }

    for(i=0 ; i<n_points ; i++)
      yfit[i] = func(x[i], b);
    
    *chisqr = calc_chisqr(y, yfit, weight, n_points, n_points-n_terms);

    if (*chisqr > chisq1) 
      *flamda *= 10;

  } while (*chisqr > chisq1);

  /* evaluate parameters and uncertainties */
  for (j=0;j<n_terms;j++)
    {
    a[j] = b[j];
    sigmaa[j] = sqrt(array[j][j]/alpha[j][j]);
    }

  *flamda /= 10;

  return(0);
}

/*----------------------------------------------------------------------*/

double  fit(double x[], double y[], double weight[],
            double (*func)(double x, double param[]),
            int n_points, int n_params,
            double param[], double sigmaparam[],
            double yfit[])
/******************************************************************\
  
  fit a function "func(x, param[])" to a given data set x[]/y[]
  with poisson weighted data points.

  Input:    x[]: array containing x-values of data points
            y[]: array containing value of data points
            weight[]: array with data point weights
            func(): user-supplied fitting function
            n_points: number of data points to fit
            n_params: number of parameters
            param[]: array containing initial parameter values

  Output:   param[]: array containing fitted parameter values
            sigmaparam[]: errors in parameters             
            yfit[]: new fit                            
            chisqr: new chisquare

  Returns:  chisqr: new chisquare
            -1: if fit doesn't converge. A reason for this could
            be that the variation of a parameter doesn't affect
            the chi^2 value. This happens if the starting 
            parameters are wrong or the fit function is independent
            of a given parameter.


\******************************************************************/
{
double chisqr, chiold, flamda;
int    i, status;
  /* lamda parameter for marquardt_fit */  
  flamda = 0.001;

  /* calculate initial fit with starting parameters */
  for (i=0 ; i<n_points ; i++)
    yfit[i] = func(x[i], param);

  /* calculate starting chi sqare */
  chisqr = calc_chisqr(y, yfit, weight, n_points, n_points-n_params);

  /* main iteration loop */
  do
    {                                 
    chiold = chisqr;
    status = marquardt_fit(x, y, weight, func, n_points, n_params,
 	                         param, sigmaparam, &flamda, 
                           yfit, &chisqr);
    if (status == -1)
      return -1;

    } while(fabs(chiold-chisqr)/chisqr > 0.001);

  return chisqr;
}

/*----------------------------------------------------------------*/

/*******************************************************************
  
    Generic fit functions

*******************************************************************/

#define FWHM_RATIO 2.35482

double f_gaussian(double x, double param[])
/* 0: peak height, 1: center, 2: FWHM */
{
double x1;
double sigma;

  /* supress division by zero */
  if (param[2] == 0) 
    return(0);       

  sigma = param[2]/FWHM_RATIO;

  x1 = (x - param[1]) / sigma;

  /* supress overflow */
  if ((x1>4) || (x1<-4)) 
    return(0);           

  /* sqare argument */
  x1 *= x1;  

  return param[0] * exp( -0.5*x1 );
}

/*----------------------------------------------------------------*/

double f_gaussexp(double x, double param[])
/* 0: peak height, 1: center, 2: FWHM, 3: split point */
{
double x1;
double sigma;

  /* supress division by zero */
  if (param[2] == 0) 
    return(0);       

  sigma = param[2]/FWHM_RATIO;

  if ( x <= param[1]+param[3] && param[3] > 0 ||
       x >= param[1]+param[3] && param[3] <= 0)
    {
    x1 = (x - param[1]) / sigma;

    /* supress overflow */
    if ((x1>4) || (x1<-4)) 
      return(0);           
    
    x1 *= x1;
  
    return param[0] * exp( -0.5*x1 );
    }
  else
    {
    x1 = param[3]/sigma/sigma*(0.5*param[3] - (x-param[1]));
    
    /* supress underflow of exp function */
    if (x1<-20) 
      return(0); 
    if (x1>20) 
      x1 = 20;     /* overflow */
    x1 = exp(x1);

    return param[0] * x1;
    }
}

/*----------------------------------------------------------------*/

double f_trapezoid(double x, double param[])
/* 0: gauss height, 1: gauss center, 2: gauss FWHM, 
   3: top left edge, 4: top right edge, 5: trap height,
   6: slope left, 7: slope right */
{ 
double x1;

  /* supress division by zero */
  if (param[2] == 0) 
    return 0;       

  /* (x-center)/sigma */
  x1 = (x - param[1]) / (param[2]/FWHM_RATIO);    

  /* supress overflow */
  if ((x1>4) || (x1<-4)) 
    return 0;           

  /* x^2 */
  x1 *= x1;  
  

  if (x >= param[3] && x <= param[4])
    return param[5] + param[0]*exp(-0.5*x1);

  if (x >= param[4] && x <= param[4] - param[5]/param[7])
    return param[7]*x + param[5] - param[7]*param[4] + param[0]*exp(-0.5*x1);

  if (x <= param[3] && x >= (param[6]*param[3]-param[5])/param[6])
    return param[6]*x + param[5] - param[6]*param[3] + param[0] * exp(-0.5*x1);

  if (x <= (param[6]*param[3]-param[5])/param[6])
    return param[0]*exp(-0.5*x1);
  
  if (x >= param[4] - param[5]/param[7])
    return param[0]*exp(-0.5*x1);
}

/*******************************************************************
  
    Gain match two histos

*******************************************************************/

/*--------- Chi^2 calculation for gain_match ----------------------*/

float calc_chisq(float *adc_histo, float *ref_histo, float norm_ref, 
                  float g, int n, float x_low, float x_high, 
                  int i_left, int i_right)
{
int    i, j;
float chisq, frac, adc, ref, delta, x;

  /* calculate chi^2 for gain g */
  chisq = 0.f;

  for (i=i_left ; i<=i_right ; i++)
    {
    /* scale adc histo with g, interpolate between bins */
    x = (float) i/n*(x_high-x_low) + x_low + 0.5f;
    x /= g;

    x = (x-x_low)/(x_high-x_low)*n;
    j = (int) x;
    frac = x - j;

    if (j>=n-1 || j<0)
      adc = 0.f;
    else
      adc = adc_histo[j] + frac * ((float) adc_histo[j+1] - adc_histo[j]);

    /* normalize events */
    adc /= g;

    ref = ref_histo[i]*norm_ref;
    delta = adc - ref;

    if (ref == 0)
      chisq += delta*delta;
    else
      chisq += delta*delta/(float)fabs(ref);
    }

  chisq /= (i_right-i_left);

  return chisq;
}

/*--------- Histogram gain evaluation -----------------------------*/

void gain_match(float *ref_histo, float *adc_histo, int n,
                float x_low, float x_high,
                float roi_low, float roi_high,
                float range_low, float range_high,
                float *gain, float *error)
/******************************************************************\
  
  Evaluate gain by matching a histo in a given region to a reference
  histo.
  
  Input:    ref_histo[]    Reference histo
            adc_histo[]    Histo which is gain matched
            x_low,x_high   Physical histo limits
            roi_low,roi_high Region where to match histos
            range_low/high Gain range to check (like 0.8..1.2)

  Output:   gain           Gain
            error          Error of gain


\******************************************************************/
{
int   i, i_low, i_high;
float sum_ref, sum_adc, norm_ref, g, g1;
float chisq, min_chisq, old_chisq;

  i_low  = (int) ((roi_low-x_low)/(x_high-x_low)*n);
  i_high = (int) ((roi_high-x_low)/(x_high-x_low)*n);

  /* normalize reference histogram */
  for (i=i_low,sum_ref=0.f,sum_adc=0.f ; i<i_high && i<n ; i++)
    {
    sum_ref += ref_histo[i];
    sum_adc += adc_histo[i];
    }

  /* don't continue for empty histos */
  if (sum_ref <= 0 || sum_adc <= 0 || n<2)
    {
    *gain = 1.f;
    *error = 1.f;
    return;
    }

  /* do range check of limits */
  if (roi_low<x_low || roi_low>=roi_high ||
      roi_high>x_high)
    {
    *gain = 1.f;
    *error = 1.f;
    return;
    }

  norm_ref = sum_adc/sum_ref;

  /* vary gain inside range in steps of 1/n */
  for (g=range_low ; g<range_high ; g+=(float) 1/n)
    {
    /* calculate chi^2 for gain g */
    chisq = calc_chisq(adc_histo, ref_histo, norm_ref, g, n, 
                       x_low, x_high, i_low, i_high);

    /* check for minimal chisq */
    if (g==range_low || chisq < min_chisq)
      {
      *gain = g;
      min_chisq = chisq;
      }
    }

  /* error estimation */
  old_chisq = min_chisq;
  for (g= *gain ; g<range_high ; g+=(float) 1/n)
    {
    /* calculate chi^2 for gain g */
    chisq = calc_chisq(adc_histo, ref_histo, norm_ref, g, n, 
                       x_low, x_high, i_low, i_high);

    /* check for chisq increase of one */
    if (chisq >= min_chisq+1)
      {
      g1 = g - (chisq - (min_chisq+1))/(chisq-old_chisq)/n;
      *error = g1 - *gain;
      break;
      }

    old_chisq = chisq;
    }
}