/********************************************************************\

  Name:         frontend.c
  Created by:   Stefan Ritt

  Contents:     Experiment specific readout code (user part) of
                Midas frontend


  Revision history
  ------------------------------------------------------------------
  date         by    modification
  ---------    ---   ------------------------------------------------
  25-JAN-95    SR    created


\********************************************************************/

/* standard include files */
#include <stdio.h>
#include <dos.h>
#include <string.h>

/* midas include file */
#include "midas.h"
#define BYTE_DEFINED

/* include files for hardware access */
#include "hardware.h"
#include "hyt1331.h"
#include "lrs1821.h"
#include "ps7106.h"
#include "lrs2373.h"
#include "lrs2365.h"
#include "pcos.h"

/* slot number etc */
#include "frontend.h"

/* trigger settings */
#define EXCL_HV
#define EXCL_ADC_CALIBRATION
#include "..\event.h"

/*-- System Globals ------------------------------------------------*/

/* The frontend name (client name) as seen by other MIDAS clients   */
char *frontend_name = "Trigger Frontend";

/* frontend_loop is called periodically if this variable is TRUE    */
BOOL frontend_call_loop = FALSE;

/* a frontend status page is displayed with this frequency in ms */
INT display_period = 3000;

/* display function in mfe.c */
extern void display(BOOL bInit);

/*-- Experiment specific Globals -----------------------------------*/

TRIGGER_SETTINGS trigger_settings;
SCALER_SETTINGS  scaler_settings;
EXP_PARAM        exp_param;

/*-- Channel mappings ----------------------------------------------*/

#define N_ADC         144   /* number of used ADC channels */
#define N_TDC         160   /* number of used TDC channels */

typedef struct {
  int slot;
  int n_channels;
} CHANNEL_MAP;

CHANNEL_MAP adc_map[] = {
  { 24, 80 },
  { 23, 64 },
  {  0  }
};

CHANNEL_MAP tdc_map[] = {
  { 10, 80 },
  {  9, 80 },
  {  0  }
};

/*-- Function declarations -----------------------------------------*/

INT  poll_trigger_event(INT count, PTYPE test);
INT  read_trigger_event(char *pevent);
INT  read_scaler_event(char *pevent);

/*-- Equipment list ------------------------------------------------*/

EQUIPMENT equipment[] = {

  { "Trigger",            /* equipment name */
    1, 0,                 /* event ID, trigger mask */
    "SYSTEM",             /* event buffer */
    EQ_POLLED,            /* equipment type */
    "MIDAS",              /* format */
    TRUE,                 /* enabled */
    RO_ODB |              /* write event to odb */
    RO_RUNNING,           /* read only when running */
    100,                  /* poll for 100ms */
    0,                    /* stop run after this event limit */
    0,                    /* don't log history */
    "", "",
    read_trigger_event,   /* readout routine */
    poll_trigger_event    /* polling routine */
  },

  { "Scaler",             /* equipment name */
    2, 0,                 /* event ID, trigger mask */
    "SYSTEM",             /* event buffer */
    EQ_PERIODIC,          /* equipment type */
    "MIDAS",              /* format */
    TRUE,                 /* enabled */
    RO_ODB |              /* write event to odb */
    RO_RUNNING |
    RO_TRANSITIONS,       /* read when running and on transitions */
    10000,                /* read every 10 sec */
    0,                    /* stop run after this event limit */
    0,                    /* don't log history */
    "", "",
    read_scaler_event     /* readout routine */
  },

  { "" }
};

/*-- ODB update callbacks ------------------------------------------*/

void trigger_update(INT hDB, INT hKey)
{
int i;

#ifdef PS7106
  // CAMAC discriminators
  CAMAC_SET_CRATE(3);
  for (i=CS_DISC_START ; i<=CS_DISC_END ; i++)
    ps7106_set(i, trigger_settings.ps7106_threshold);
#endif

#ifdef LRS2365
  // LRS 2365
  CAMAC_SET_CRATE(1);
  lrs2365_set(CS_LRS2365, trigger_settings.lrs_2365_coeff);
#endif

#ifdef LRS4518
  // LRS4518 ECL delay
  CAMAC_SET_CRATE(1);
  for (i=0 ; i<16 ; i++)            // Mod 100: 2ns steps
    CAMO(CS_LRS4518_1, i, 16, trigger_settings.lrs_4518_1[i]/2);
  for (i=0 ; i<16 ; i++)            // Mod 300: 8ns steps
    CAMO(CS_LRS4518_2, i, 16, trigger_settings.lrs_4518_2[i]/8);
#endif

#ifdef DA500
  // DA500 DAC
  CAMO(CS_DACHI, 0, 26, 0); // enable auto mode
  CAMO(CS_DACHI, 1, 16, 1); // x1
  CAMO(CS_DACHI, 0, 16, (int) (trigger_settings.da500_hi*100+0.5));

  CAMO(CS_DACLO, 0, 26, 0); // enable auto mode
  CAMO(CS_DACLO, 1, 16, 1); // x1
  CAMO(CS_DACLO, 0, 16, (int) (trigger_settings.da500_lo*100+0.5));
#endif

  // IO506 output pattern
  CAMO8(CS_IO2, 1, 16, trigger_settings.io506);

  if (hDB != 0)
    {
    equipment[0].odb_in++;
    display(TRUE);
    }
}

/*------------------------------------------------------------------*/

/********************************************************************\
              Callback routines for system transitions

  These routines are called whenever a system transition like start/
  stop of a run occurs. The routines are called on the following
  occations:

  frontend_init:  When the frontend program is started. This routine
                  should initialize the hardware.

  frontend_exit:  When the frontend program is shut down. Can be used
                  to releas any locked resources like memory, commu-
                  nications ports etc.

  begin_of_run:   When a new run is started. Clear scalers, open
                  rungates, etc.

  end_of_run:     Called on a request to stop a run. Can send final
                  scalers, end-of-run event, and close run gates.

  pause_run:      When a run is paused. Should disable trigger events.

  resume_run:     When a run is resumed. Should enable trigger events.

\********************************************************************/

/*-- Frontend Init -------------------------------------------------*/

INT frontend_init()
{
HNDLE hDB, hKey, hKeyRoot;
int   i, slot;
char  filename[80];
DWORD csr;
INT   size;
EXP_PARAM_STR(exp_param_str);
TRIGGER_SETTINGS_STR(trigger_settings_str);
SCALER_SETTINGS_STR(scaler_settings_str);

  // Open MIDAS online database
  cm_get_experiment_database(&hDB, &hKey);

  // experiment parameters
  db_create_record(hDB, 0, "/Experiment/Run Parameters", strcomb(exp_param_str));
  db_find_key(hDB, 0, "/Experiment/Run Parameters", &hKey);
  if (db_open_record(hDB, hKey, &exp_param,
        sizeof(exp_param), MODE_READ, NULL) != DB_SUCCESS)
    {
    cm_msg(MERROR, "frontend_init", "Cannot open Run Parameters in ODB");
    return -1;
    }

  // trigger settings
  db_create_record(hDB, 0, "/Equipment/Trigger/Settings", strcomb(trigger_settings_str));
  db_find_key(hDB, 0, "/Equipment/Trigger/Settings", &hKey);
  if (db_open_record(hDB, hKey, &trigger_settings,
        sizeof(trigger_settings), MODE_READ, trigger_update) != DB_SUCCESS)
    {
    cm_msg(MERROR, "frontend_init", "Cannot open Trigger Settings in ODB");
    return -1;
    }

  // scaler settings
  db_create_record(hDB, 0, "/Equipment/Scaler/Settings", strcomb(scaler_settings_str));
  db_find_key(hDB, 0, "/Equipment/Scaler/Settings", &hKey);
  if (db_open_record(hDB, hKey, &scaler_settings,
        sizeof(scaler_settings), MODE_READ, NULL) != DB_SUCCESS)
    {
    cm_msg(MERROR, "frontend_init", "Cannot open Scaler Settings in ODB");
    return -1;
    }

  /*-- CAMAC Initialization ----------------------------------------*/

  // Crate 0
  CAMAC_SET_CRATE(0);
  CAMAC_Z_CYCLE();
  CAMAC_C_CYCLE();

  // System crate (1)
  CAMAC_SET_CRATE(1);
  CAMAC_Z_CYCLE();
  CAMAC_C_CYCLE();

  // Crate 2
  CAMAC_SET_CRATE(2);
  CAMAC_Z_CYCLE();
  CAMAC_C_CYCLE();

  // Crate 3
  CAMAC_SET_CRATE(3);
  CAMAC_Z_CYCLE();
  CAMAC_C_CYCLE();

  // init IO 506
  CAMAC_SET_CRATE(1);
  CAMO8(CS_IO, 1, 26, 0); // enable auto mode
  CAMO8(CS_IO, 2, 16, 0); // NIM level, int. clock
  CAMO8(CS_IO, 0, 26, 0); // enable LAM Flip-Flop

  // Initialize trigger
  trigger_update(0, 0);

#ifdef PCOS
  CAMAC_SET_CRATE(1);
  pcos_init(CS_DBI);
#endif

  /*-- FASTBUS Initialization --------------------------------------*/

  if (!fb_init())
    return -1;

  fb_exec(FPREQ);  // request mastership

#ifdef LRS1882
  for (i=0 ; adc_map[i].slot > 0 ; i++)
    {
    slot = adc_map[i].slot;

    /**** check and reset ADCs ****/
    fb_fpac(slot);
    fb_fpr(&csr);
    if (csr != 0x10440300)
      {
      cm_msg(MERROR, "frontend_init", "LRS 1882F not found in slot %d (csr=%lX)\n", slot, csr);
      return -1;
      }

    fb_fpwnta(0);        // CSR 0
    fb_fpw(0x00000800L); // internal gate
    fb_fpw(0x08000000L); // clear
    fb_fpw(0x40000000L); // reset

    // test if in N mode
    fb_fpw(0x00000800L); // internal gate
    delay(1);            // wait 1 ms
    fb_fpwnta(1);        // CSR 1
    fb_fpr(&csr);
    if (csr != 0x60000000)
      {
      cm_msg(MERROR, "frontend_init", "LRS 1882F not in N mode in slot %d\n", slot);
      return -1;
      }
    }
#endif

#ifdef LRS1877
  for (i=0 ; tdc_map[i].slot > 0 ; i++)
    {
    slot = tdc_map[i].slot;

    // check for module
    fb_fpac(slot);
    fb_fpwnta(0); /* CSR0 */
    fb_fpr(&csr);
    if (csr != 0x103C0000l)
      {
      cm_msg(MERROR, "frontend_init", "No LRS 1877 found in slot %d. (data=0x%X)\n",
             slot, csr);
      return -1;
      }

    // reset module
    fb_fpw (0x40000000L);

    // set common start, leading + trailing edge enable, 512ns TO
    fb_fpwnta(1); // CSR1
    fb_fpw (0xE0000040L);

    fb_fpwnta(18); // CSR18
    fb_fpw (0x0000FFF0L); /* set full scale to 32us, LIFO depth = 16 */
    }
#endif

#ifdef PS10C6
  fb_fpac(FBSLOT_TDC1);

  fb_fpwnta(0);
  /* set CSR0:
  Disable pedestals, disable mega block, disable upper and lower thresh,
  disable SRs, reset data available, data protect disabled
  */
  fb_fpw(0x3BB00044L);  /* Like MEGA, but with data protect enabled */

  // Phillips TDC #2
  fb_fpac(FBSLOT_TDC2);

  fb_fpwnta(0);
  /* set CSR0:
  Disable pedestals, disable mega block, disable upper and lower thresh,
  disable SRs, reset data available, data protect disabled
  */
  fb_fpw(0x3BB00044L);  /* Like MEGA, but with data protect enabled */
#endif

  /*----------------------------------------------------------------*/

  CAMAC_SET_CRATE(1);

  // clear LAM Flip-Flop
  CAMO(CS_IO, 0, 10, 0);

  // reset LAM logic
  CAMO8(CS_IO, 1, 16, 0xF);
  CAMO8(CS_IO, 1, 16, 0);

  return CM_SUCCESS;
}

/*-- Frontend Exit -------------------------------------------------*/

INT frontend_exit()
{
  return CM_SUCCESS;
}

/*-- Begin of Run --------------------------------------------------*/

INT begin_of_run(INT run_number, char *error)
{
int i, slot;

  // clear scalers
  CAMAC_SET_CRATE(1);
  CAMAC_Z_CYCLE();
  CAMAC_SET_CRATE(2);
  CAMAC_Z_CYCLE();

  // init IO 506
  CAMAC_SET_CRATE(1);
  CAMO8(CS_IO, 1, 26, 0); // enable auto mode
  CAMO8(CS_IO, 2, 16, 0); // NIM level, int. clock
  CAMO8(CS_IO, 0, 26, 0); // enable LAM Flip-Flop

  // Initialize trigger
  trigger_update(0, 0);

#ifdef LRS1877
  // clear TDCs
  for (i=0 ; tdc_map[i].slot > 0 ; i++)
    {
    slot = tdc_map[i].slot;

    // clear TDC
    fb_fpac(slot);
    fb_fpwnta(16);
    fb_fpw(0x00000700);  // set read buffer to 07, write buffer to 0
    }
#endif

#ifdef LRS1882
  // clear ADCs
  for (i=0 ; adc_map[i].slot > 0 ; i++)
    {
    slot = adc_map[i].slot;

    // clear ADC
    fb_fpac(slot);
    fb_fpwnta(0);        // CSR 0
    fb_fpw(0x08000000L); // clear
    }
#endif

  // disconnect, by addressing 1821
  fb_fpac(1);

  // clear LAM Flip-Flop
  CAMO(CS_IO, 0, 10, 0);

  // reset LAM logic
  CAMO8(CS_IO, 1, 16, 0xF);
  CAMO8(CS_IO, 1, 16, 0);

  return CM_SUCCESS;
}

/*-- End of Run ----------------------------------------------------*/

INT end_of_run(INT run_number, char *error)
{
  return CM_SUCCESS;
}

/*-- Pause Run -----------------------------------------------------*/

INT pause_run(INT run_number, char *error)
{
  return CM_SUCCESS;
}

/*-- Resuem Run ----------------------------------------------------*/

INT resume_run(INT run_number, char *error)
{
  return CM_SUCCESS;
}

/*-- Frontend Loop -------------------------------------------------*/

INT frontend_loop()
{
  return CM_SUCCESS;
}

/*------------------------------------------------------------------*/

/********************************************************************\

  Readout routines for different events

\********************************************************************/

/*-- Poll Trigger event --------------------------------------------*/

INT poll_trigger_event(INT count, PTYPE test)
{
INT  i;
BYTE q;

  CAMIQ_RSTART(CS_IO, 0, 8);
  for (i=0 ; i<count ; i++)
    {
    CAMIQ_REPEAT(&q);
    if (q)
      if (!test)
        return TRUE;
    }

  return FALSE;
}

/*-- Read trigger event --------------------------------------------*/

INT read_trigger_event(char *pevent)
{
DWORD          *pdata;
WORD           *pdataw;
int            i, j, n_tdc, slot;
DWORD          csr;
WORD           count, *padc, *parray;
WORD           data;
BYTE           q, x;
LRS1877_HEADER *lrs1877_header;
LRS1877_DATA   *lrs1877_data;

  // signal start of readout
  CAMAC_SET_CRATE(1);
  CAMO8(CS_IO, 1, 16, 1<<7);

  // init bank structure
  bk_init(pevent);

#ifdef CAMAC_COINC
  bk_create(pevent, "TRIG", TID_WORD, &pdataw);

  CAMI(CS_COINC, 0, 2, pdataw++);
  bk_close(pevent, pdataw);
#endif

#ifdef PCOS
  bk_create(pevent, "PCOS", TID_WORD, &pdataw);
  i = pcos_read(CS_DBI, pdataw);
  bk_close(pevent, pdataw+i);
#endif

#ifdef CAMAC_ADCS
  bk_create(pevent, "MADC", TID_WORD, &pdataw);
  CAMI_LOOP(CS_ADC0, 0, 3, 0,  pdataw += 4);       // readout
  CAMO8(CS_ADC0, 0, 9, 0);                         // clear
  bk_close(pevent, pdataw);
#endif

#ifdef CAMAC_TDCS
  bk_create(pevent, "MTDC", TID_WORD, &pdataw);
  CAMI_LOOP(CS_TDC0, 0, 3, 0,  pdataw += 4);       // readout
  CAMO8(CS_TDC0, 0, 9, 0);                         // clear
  bk_close(pevent, pdataw);
#endif

#ifdef LRS1877
  bk_create(pevent, "LTDC", TID_DWORD | TID_LRS1877, &pdata);

  // loop over TDCs
  for (i=0 ; tdc_map[i].slot > 0 ; i++)
    {
    slot = tdc_map[i].slot;

    // Load next event (transfer data from MTD133 to 1877 memory)
    fb_fpac(slot);
    fb_fpwnta(0);
    fb_fpw(0x00000400L);

    // Non-block transfer (problems in 96 beam time)
    fb_fpad(slot);
    fb_fpr(pdata);
    lrs1877_header = (LRS1877_HEADER *) pdata;
    n_tdc = lrs1877_header->count-1;
    if (n_tdc > 96)
      n_tdc = 96;

    for (j=0 ; j<n_tdc ; j++)
      fb_fpr(pdata++);

    /*
    // Execute block transfer
    fb_fpad(slot);
    count = 2 * (N_TDC+1);
    fb_fprbl(pdata, &count, 0);
    pdata += count/2;
    */

    // clear TDC
    fb_fpac(slot);
    fb_fpwnta(16);
    fb_fpw(0x00000700);  // set read buffer to 07, write buffer to 0
    }

  bk_close(pevent, pdata);
#endif

#ifdef PS10C6
  bk_create(pevent, "PTDC", TID_DWORD, &pdata);

  fb_fpad(FBSLOT_TDC0);
  fb_fpwnta(1); // DSR1

  count = 34;
  fb_fprbl(pdata, &count, 0);
  pdata += count/2;
  bk_close(pevent, pdata);
#endif

#ifdef LRS1882
  bk_create(pevent, "FADC", TID_DWORD | TID_LRS1882, &pdata);

  // wait for CIP signal going off, must be connected over IN1 of SMI
  do
    {
    i = fb_in1();
    } while(i);

  // loop over ADCs
  for (i=0 ; adc_map[i].slot > 0 ; i++)
    {
    slot = adc_map[i].slot;

    // Execute block transfer
    fb_fpad(slot);

    count = 2 * adc_map[i].n_channels;

    fb_fprbl((WORD *) pdata, &count, 0);
    pdata += count/2;

    // clear ADC
    fb_fpac(slot);
    fb_fpwnta(0);        // CSR 0
    fb_fpw(0x08000000L); // clear
    }

  bk_close(pevent, pdata);
#endif

  fb_fpac(1); // disconnect, by addressing 1821

  // clear LAM Flip-Flop
  CAMO(CS_IO, 0, 10, 0);

  // reset LAM logic
  CAMO8(CS_IO, 1, 16, 0xF);
  CAMO8(CS_IO, 1, 16, 0);

  // return event size
  return bk_size(pevent);
}

/*-- Scaler event --------------------------------------------------*/

INT read_scaler_event(char *pevent)
{
DWORD       *pdata;
int         i, slot;

  // init bank structure
  bk_init(pevent);

#ifdef CAMAC_SCALERS
  bk_create(pevent, "NIMS", TID_DWORD, &pdata);
  CAMAC_SET_CRATE(1);
  for (i=0 ; i<12 ; i++)
    CAMI24(CS_SCALER0, i, 2, pdata++);
  bk_close(pevent, pdata);
#endif

#ifdef PS7132
  bk_create(pevent, "SCLR", TID_DWORD, &pdata);
  CAMAC_SET_CRATE(2);
  for (slot=CS_SCLR_START ; slot<=CS_SCLR_END ; slot++)
    {
    CAMO(slot, 1, 17, 0); // select scaler bank 0-15
    for (i=0 ; i<32 ; i++)
      {
      if (i==16)
        CAMO(slot, 1, 17, 1); // switch to scaler bank 16-31

      CAMI24(slot, i & 0xF, 2, pdata);
      pdata++;
      }
    }
  bk_close(pevent, pdata);
  CAMAC_SET_CRATE(1);
#endif

  return bk_size(pevent);
}
